"""
.. module:: framework
    :platform: Linux
    :synopsis: framework definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import List

from libkirk.com import ComChannel
from libkirk.data import (
    Suite,
    Test,
)
from libkirk.results import TestResults


class Framework:
    """
    Framework definition. Implement this class if you need to support more
    testing frameworks inside the application.
    """

    async def get_suites(self, channel: ComChannel) -> List[str]:
        """
        Return the list of available suites.

        :param channel: Communication channel.
        :type channel: ComChannel
        :return: List of suites names.
        :rtype: list(str)
        """
        raise NotImplementedError()

    async def find_command(self, channel: ComChannel, command: str) -> Test:
        """
        Search for command inside Framework folder and, if it's not found,
        search for command in the SUT. Then return a Test object which can be
        used to execute command.

        :param channel: Communication channel.
        :type channel: ComChannel
        :param command: Command to execute.
        :type command: str
        :return: Test object that has been found.
        :rtype: Test
        """
        raise NotImplementedError()

    async def find_suite(self, channel: ComChannel, name: str) -> Suite:
        """
        Search for suite with given name inside the SUT.

        :param channel: Communication channel.
        :type channel: ComChannel
        :param suite: Name of the suite.
        :type suite: str
        :return: Suite object that has been found.
        :rtype: Suite
        """
        raise NotImplementedError()

    async def read_result(
        self, test: Test, stdout: str, retcode: int, exec_t: float
    ) -> TestResults:
        """
        Return test results accoding with runner output and Test definition.

        :param test: Test definition object.
        :type test: Test
        :param stdout: Test stdout.
        :type stdout: str
        :param retcode: Test return code.
        :type retcode: int
        :param exec_t: Test execution time in seconds.
        :type exec_t: float
        :return: Test results.
        :rtype: TestResults
        """
        raise NotImplementedError()



"""
.. module:: ltx
    :platform: Linux
    :synopsis: module containing LTX communication class

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import logging
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
)

import libkirk
from libkirk.errors import LTXError
from libkirk.io import AsyncFile

try:
    import msgpack
except ModuleNotFoundError:
    pass


class Request:
    """
    LTX client request.
    """

    ERROR = 0xFF
    VERSION = 0x00
    PING = 0x01
    PONG = 0x02
    GET_FILE = 0x03
    SET_FILE = 0x04
    ENV = 0x05
    CWD = 0x06
    EXEC = 0x07
    RESULT = 0x08
    LOG = 0x09
    DATA = 0xA0
    KILL = 0xA1
    MAX_SLOTS = 127
    ALL_SLOTS = 128
    MAX_ENVS = 16

    def __init__(self) -> None:
        self._logger = logging.getLogger("ltx.request")
        self._completed = False
        self._done_coro = []

    @property
    def completed(self) -> bool:
        """
        :return: True if request has been completed. False otherwise.
        :rtype: bool
        """
        return self._completed

    def add_done_coro(self, coro: Callable) -> None:
        """
        Add done event to request.

        :param coro: Called when request is completed.
        :type coro: Callable
        """
        self._done_coro.append(coro)

    async def _raise_complete(self, *args: Any) -> None:
        """
        Raise the complete callback with given data.
        """
        if self._done_coro:
            self._logger.info("Raising 'on_complete(self, %s)'", args)

            for coro in self._done_coro:
                await coro(self, *args)

        self._completed = True

    async def pack(self) -> bytes:
        """
        Pack LTX request into bytes.

        :return: Request bytes.
        :rtype: bytes
        """
        raise NotImplementedError()

    async def feed(self, message: List[Any]) -> None:
        """
        Feed request queue with data and return when the request
        has been completed.

        :param message: Processed msgpack message.
        :type message: list
        """
        raise NotImplementedError()


class Requests:
    """
    Class container for LTX requests.
    """

    class version(Request):
        """
        VERSION request.
        """

        async def pack(self) -> bytes:
            pkg = msgpack.packb([self.VERSION])
            if not pkg:
                raise LTXError("Can't pack VERSION request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if message[0] == self.VERSION:
                ver = message[1]

                self._logger.debug("version=%s", ver)
                await self._raise_complete(ver)

    class ping(Request):
        """
        PING request.
        """

        def __init__(self) -> None:
            super().__init__()
            self._echoed = False

        async def pack(self) -> bytes:
            pkg = msgpack.packb([self.PING])
            if not pkg:
                raise LTXError("Can't pack PING request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if message[0] == self.PING:
                self._logger.info("PING echoed back")
                self._logger.info("Waiting for PONG")
                self._echoed = True
            elif message[0] == self.PONG:
                if not self._echoed:
                    raise LTXError("PONG received without PING echo")

                end_t = message[1]

                self._logger.debug("end_t=%s", end_t)
                await self._raise_complete(end_t)

    class env(Request):
        """
        ENV request.
        """

        def __init__(self, slot_id: int, key: str, value: str) -> None:
            """
            :param slot_id: Command table ID. Can be None if we want to apply
                the same environment variable to all executions.
            :type slot_id: int
            :param key: Key of the environment variable.
            :type key: str
            :param value: Value of the environment variable.
            :type value: str
            """
            super().__init__()

            if slot_id and (slot_id < 0 or slot_id > self.ALL_SLOTS):
                raise ValueError(f"Out of bounds slot ID [0-{self.ALL_SLOTS}]")

            if not key:
                raise ValueError("key is empty")

            if not value:
                raise ValueError("value is empty")

            self._slot_id = self.ALL_SLOTS if slot_id is None else slot_id
            self._key = key
            self._value = value

        async def pack(self) -> bytes:
            pkg = msgpack.packb([self.ENV, self._slot_id, self._key, self._value])

            if not pkg:
                raise LTXError("Can't pack ENV request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if len(message) > 1 and message[1] != self._slot_id:
                return

            if message[0] == self.ENV:
                self._logger.info("ENV echoed back")
                await self._raise_complete(self._slot_id, self._key, self._value)

    class cwd(Request):
        """
        CWD request.
        """

        def __init__(self, slot_id: int, path: str) -> None:
            """
            :param slot_id: Command table ID. Can be None if we want to apply
                the same current working directory to all executions
            :type slot_id: int
            :param path: Current working path.
            :type path: str
            """
            super().__init__()

            if slot_id is not None and (slot_id < 0 or slot_id > self.ALL_SLOTS):
                raise ValueError(f"Out of bounds slot ID [0-{self.ALL_SLOTS}]")

            if not path:
                raise ValueError("path is empty")

            self._slot_id = self.ALL_SLOTS if slot_id is None else slot_id
            self._path = path

        async def pack(self) -> bytes:
            pkg = msgpack.packb(
                [
                    self.CWD,
                    self._slot_id,
                    self._path,
                ]
            )

            if not pkg:
                raise LTXError("Can't pack CWD request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if len(message) > 1 and message[1] != self._slot_id:
                return

            if message[0] == self.CWD:
                self._logger.info("CWD echoed back")
                await self._raise_complete(self._slot_id, self._path)

    class get_file(Request):
        """
        GET_FILE request.
        """

        def __init__(self, path: str) -> None:
            """
            :param path: Path of the file to read.
            :type path: str
            """
            super().__init__()

            if not path:
                raise ValueError("path is empty")

            self._path = path
            self._data = []

        async def pack(self) -> bytes:
            pkg = msgpack.packb(
                [
                    self.GET_FILE,
                    self._path,
                ]
            )

            if not pkg:
                raise LTXError("Can't pack GET_FILE request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if message[0] == self.DATA:
                self._logger.info("Data received")
                self._data.append(message[1])
            elif message[0] == self.GET_FILE:
                self._logger.info("GET_FILE echoed back")
                await self._raise_complete(self._path, b"".join(self._data))

    class set_file(Request):
        """
        SET_FILE request.
        """

        def __init__(self, path: str, data: bytes) -> None:
            """
            :param path: path of the file to write
            :type path: str
            :param data: data to write on file
            :type data: bytes
            """
            super().__init__()

            if not path:
                raise ValueError("path is empty")

            if not data:
                raise ValueError("data is empty")

            self._path = path
            self._data = data

        async def pack(self) -> bytes:
            pkg = msgpack.packb(
                [
                    self.SET_FILE,
                    self._path,
                    self._data,
                ]
            )

            if not pkg:
                raise LTXError("Can't pack SET_FILE request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if message[0] == self.SET_FILE and message[1] == self._path:
                self._logger.info("SET_FILE echoed back")
                await self._raise_complete(self._path, self._data)

    class execute(Request):
        """
        EXEC request.
        """

        def __init__(
            self, slot_id: int, command: str, stdout_coro: Optional[Callable] = None
        ) -> None:
            """
            :param slot_id: command table ID
            :type slot_id: int
            :param command: command to run
            :type command: str
            :param stdout_coro: called when new data arrives in stdout
            :type stdout_coro: Callable
            """
            super().__init__()

            if slot_id is None:
                raise ValueError("slot_id is empty")

            if slot_id < 0 or slot_id > self.MAX_SLOTS:
                raise ValueError(f"Out of bounds slot ID [0-{self.MAX_SLOTS}]")

            if not command:
                raise ValueError("Command is empty")

            self._slot_id = slot_id
            self._command = command
            self._stdout_coro = stdout_coro
            self._stdout = []
            self._echoed = False

        async def pack(self) -> bytes:
            pkg = msgpack.packb(
                [
                    self.EXEC,
                    self._slot_id,
                    self._command,
                ]
            )

            if not pkg:
                raise LTXError("Can't pack EXEC request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if len(message) > 1 and message[1] != self._slot_id:
                return

            if message[0] == self.EXEC:
                self._logger.info("EXEC echoed back")
                self._echoed = True
            elif message[0] == self.LOG:
                if not self._echoed:
                    raise LTXError("LOG received without EXEC echo")

                log = message[3]

                if log:
                    self._logger.info("LOG replied with data: %s", repr(log))
                    self._stdout.append(log)

                    if self._stdout_coro:
                        await self._stdout_coro(log)
            elif message[0] == self.RESULT:
                if not self._echoed:
                    raise LTXError("RESULT received without EXEC echo")

                self._logger.info("RESULT received")

                stdout = "".join(self._stdout)
                time_ns = message[2]
                si_code = message[3]
                si_status = message[4]

                self._logger.debug(
                    "time_ns=%s, si_code=%s, si_status=%s", time_ns, si_code, si_status
                )

                await self._raise_complete(time_ns, si_code, si_status, stdout)

    class kill(Request):
        """
        KILL request.
        """

        def __init__(self, slot_id: int) -> None:
            """
            :param slot_id: command table ID
            :type slot_id: int
            """
            super().__init__()

            if slot_id is None:
                raise ValueError("slot_id is empty")

            if slot_id < 0 or slot_id > self.MAX_SLOTS:
                raise ValueError(f"Out of bounds slot ID [0-{self.MAX_SLOTS}]")

            self._slot_id = slot_id

        async def pack(self) -> bytes:
            pkg = msgpack.packb(
                [
                    self.KILL,
                    self._slot_id,
                ]
            )

            if not pkg:
                raise LTXError("Can't pack KILL request")

            return pkg

        async def feed(self, message: List[Any]) -> None:
            if self.completed:
                return

            if len(message) > 1 and message[1] != self._slot_id:
                return

            if message[0] == self.KILL:
                self._logger.info("KILL echoed back")
                await self._raise_complete(self._slot_id)


class LTX:
    """
    This class communicates with LTX by processing given requests.
    Typical usage is the following:

    .. code-block:: python

        async with LTX(infile, outfile) as ltx:
            # create requests
            request1 = Requests.execute("echo 'hello world' > myfile")
            request2 = Requests.get_file("myfile")

            # set the complete event
            request1.add_done_coro(exec_complete_handler)
            request2.add_done_coro(get_file_complete_handler)

            # send request
            ltx.send([request1, request2])

            # process events output
            ...

    """

    BUFFSIZE = 1 << 21

    def __init__(self, infile: str, outfile: str) -> None:
        self._logger = logging.getLogger("ltx")
        self._requests = []
        self._stop = False
        self._infile = infile
        self._outfile = outfile
        self._lock = asyncio.Lock()
        self._task: Optional[asyncio.Task] = None
        self._messages = asyncio.Queue()
        self._exception: Optional[Exception] = None

    async def __aenter__(self) -> Any:
        """
        Connect to the LTX service.
        """
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        """
        Disconnect from LTX service.
        """
        await self.disconnect()
        return True

    @property
    def connected(self) -> bool:
        """
        True if connected, False otherwise.
        """
        if not self._task:
            return False

        return not self._task.done()

    async def connect(self) -> None:
        """
        Connect to LTX.
        """
        if self.connected:
            return

        self._logger.info("Connecting to LTX")

        self._exception = None
        self._task = libkirk.create_task(self._polling())

        if self._exception:
            raise self._exception

        self._logger.info("Connected")

    async def disconnect(self) -> None:
        """
        Disconnect from LTX service.
        """
        if not self.connected:
            return

        self._logger.info("Disconnecting")
        self._stop = True

        # send the stop message
        await self._messages.put(None)

        while self.connected:
            await asyncio.sleep(0.005)
            if self._exception:
                raise self._exception

        if self._exception:
            raise self._exception

        self._logger.info("Disconnected")

    async def send(self, requests: List[Request]) -> None:
        """
        Send requests to LTX service. The order is preserved during
        requests execution.

        :param requests: List of requests to send.
        :type requests: list
        """
        if not requests:
            raise ValueError("No requests given")

        if not self.connected:
            raise LTXError("Client is not connected to LTX")

        async with self._lock:
            self._logger.info("Sending requests")
            self._requests.extend(requests)

            data = [await req.pack() for req in requests]
            tosend = b"".join(data)

            async with AsyncFile(self._infile, "wb") as afile:
                await afile.write(tosend)

    async def gather(self, requests: List[Request]) -> Dict[Request, Any]:
        """
        Gather multiple requests and wait for the response, then return all
        rquests' replies inside a dictionary that maps requests with their
        reply.

        :param requests: List of requests to process.
        :type requests: list(Request)
        :return: Dictionary containing Request as keys and data from completed
            requests as values.
        :rtype: dict
        """
        req_len = len(requests)
        replies = {}

        async def on_complete(req: Request, *args: List) -> None:
            replies[req] = args

        for req in requests:
            req.add_done_coro(on_complete)

        await self.send(requests)

        while len(replies) != req_len:
            await asyncio.sleep(0.005)
            if self._exception:
                raise self._exception

        if self._exception:
            raise self._exception

        return replies

    async def _polling(self) -> None:
        """
        Read and process messages coming from LTX.
        """
        self._logger.info("Starting producer")

        with open(self._outfile, "rb", buffering=0) as afile:
            # force utf-8 encoding by using raw=False
            unpacker = msgpack.Unpacker(raw=False)

            def _read() -> Any:
                """
                Read the last available data.
                """
                data = afile.read(self.BUFFSIZE)
                self._messages.put_nowait(data)

            loop = libkirk.get_event_loop()
            # pyrefly: ignore[bad-argument-type]
            loop.add_reader(afile.fileno(), _read)

            try:
                while not self._stop:
                    data = await self._messages.get()
                    if not data:
                        break

                    self._logger.debug("Unpacking bytes: %s", data)

                    unpacker.feed(data)

                    while True:
                        try:
                            msg = unpacker.unpack()
                            if not msg:
                                continue

                            self._logger.info("Received message: %s", msg)
                            if not isinstance(msg, list):
                                raise LTXError("Message must be an array")

                            if msg[0] == Request.ERROR:
                                raise LTXError(msg[1])

                            await self._feed_requests(msg)
                        except msgpack.OutOfData:
                            break
            except LTXError as err:
                loop.remove_reader(afile.fileno())
                self._exception = err
            finally:
                self._logger.info("Producer has stopped")

    async def _feed_requests(self, data: List[Request]) -> None:
        """
        Feed the list of requests with given data.
        """
        pos = 0
        while pos < len(self._requests):
            request = self._requests[pos]
            await request.feed(data)

            if request.completed:
                del self._requests[pos]
            else:
                pos += 1



"""
.. module:: shell
    :platform: Linux
    :synopsis: module containing shell communication channel

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import contextlib
import logging
import os
import signal
import time
from asyncio.subprocess import Process
from typing import (
    Any,
    Dict,
    Optional,
)

from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import (
    CommunicationError,
    KernelPanicError,
)
from libkirk.io import AsyncFile


class ShellComChannel(ComChannel):
    """
    Channel implementing host's shell communication.
    """

    BUFFSIZE = 1024

    _name = "shell"

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.shell")
        self._fetch_lock = asyncio.Lock()
        self._procs = []
        self._active = False
        self._stop = False

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        pass

    @property
    def config_help(self) -> Dict[str, str]:
        # cwd and env are given by default, so no options are needed
        return {}

    @property
    def parallel_execution(self) -> bool:
        return True

    async def active(self) -> bool:
        return self._active

    @staticmethod
    async def _process_alive(proc: Process) -> bool:
        """
        Return True if process is alive and running.
        """
        with contextlib.suppress(asyncio.TimeoutError):
            returncode = await asyncio.wait_for(proc.wait(), 1e-6)
            if returncode is not None:
                return False

        return True

    async def _kill_process(self, proc: Process) -> None:
        """
        Kill a process and all its subprocesses.
        """
        self._logger.info("Kill process %d", proc.pid)

        try:
            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)
        except ProcessLookupError:
            # process has been killed already
            pass

    async def ping(self) -> float:
        if not await self.active():
            raise CommunicationError("Shell is not running")

        ret = await self.run_command("test .")
        if not ret:
            raise CommunicationError("'test' command failed in shell")

        reply_t = ret["exec_time"]

        return reply_t

    async def communicate(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if await self.active():
            raise CommunicationError("Shell is running")

        self._active = True

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not await self.active():
            return

        self._logger.info("Stopping shell communication")
        self._stop = True

        try:
            if self._procs:
                self._logger.info("Terminating %d process(es)", len(self._procs))

                for proc in self._procs:
                    await self._kill_process(proc)

                await asyncio.gather(*[proc.wait() for proc in self._procs])

                self._logger.info("Process(es) terminated")

            if self._fetch_lock.locked():
                self._logger.info("Terminating data fetch")

                async with self._fetch_lock:
                    pass
        finally:
            self._stop = False
            self._active = False
            self._logger.info("Shell communication has stopped")

    async def run_command(
        self,
        command: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        iobuffer: Optional[IOBuffer] = None,
    ) -> Optional[Dict[str, Any]]:
        if not command:
            raise ValueError("command is empty")

        if not await self.active():
            raise CommunicationError("Shell is not running")

        self._logger.info("Executing command: '%s'", command)

        ret = None
        proc = None
        t_end = 0
        stdout = ""

        try:
            kwargs: Dict[str, Any] = {
                "stdout": asyncio.subprocess.PIPE,
                "stderr": asyncio.subprocess.STDOUT,
                "cwd": cwd if cwd else None,
                "start_new_session": True,
            }

            if env:
                # commands can fail when env is defined, so we just skip
                # env usage if dictionary is empty
                kwargs["env"] = env

            proc = await asyncio.create_subprocess_shell(command, **kwargs)
            if not proc or not proc.stdout:
                raise CommunicationError(f"Can't create any subprocess for '{command}'")

            self._procs.append(proc)

            t_start = time.time()
            panic = False

            while True:
                line = await proc.stdout.read(self.BUFFSIZE)
                sline = line.decode(encoding="utf-8", errors="ignore")

                if iobuffer:
                    await iobuffer.write(sline)

                stdout += sline
                panic = "Kernel panic" in stdout[-2 * self.BUFFSIZE :]

                if not await self._process_alive(proc):
                    break

            await proc.wait()

            t_end = time.time() - t_start

            if panic:
                raise KernelPanicError()
        finally:
            if proc:
                self._procs.remove(proc)

                await self._kill_process(proc)
                await proc.wait()

                ret = {
                    "command": command,
                    "stdout": stdout,
                    "returncode": proc.returncode
                    if proc.returncode is not None
                    else -1,
                    "exec_time": t_end,
                }

                self._logger.debug("return data=%s", ret)

        self._logger.info("Command executed")

        return ret

    # pyrefly: ignore[bad-return]
    async def fetch_file(self, target_path: str) -> bytes:
        if not target_path:
            raise ValueError("target path is empty")

        if not os.path.isfile(target_path):
            raise CommunicationError(f"'{target_path}' file doesn't exist")

        if not await self.active():
            raise CommunicationError("Shell is not running")

        async with self._fetch_lock:
            self._logger.info("Downloading '%s'", target_path)

            try:
                async with AsyncFile(target_path, "rb") as ftarget:
                    return await ftarget.read()
            except IOError as err:
                raise CommunicationError(err) from err
            finally:
                self._logger.info("File copied")



"""
.. module:: export
    :platform: Linux
    :synopsis: module containing exporters definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import json
import logging
import os
from typing import List

from libkirk.errors import ExporterError
from libkirk.io import AsyncFile
from libkirk.results import (
    ResultStatus,
    SuiteResults,
)


class Exporter:
    """
    A class used to export Results into report file.
    """

    async def save_file(self, results: List[SuiteResults], path: str) -> None:
        """
        Save report into a file by taking information from SUT and testing
        results.

        :param results: List of suite results to export.
        :type results: list(SuiteResults)
        :param path: Path of the file to save.
        :type path: str
        """
        raise NotImplementedError()


class JSONExporter(Exporter):
    """
    Export testing results into a JSON file.
    """

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.json")

    async def save_file(self, results: List[SuiteResults], path: str) -> None:
        if not results or len(results) == 0:
            raise ValueError("results is empty")

        if not path:
            raise ValueError("path is empty")

        if os.path.exists(path):
            raise ExporterError(f"'{path}' already exists")

        self._logger.info("Exporting JSON report into %s", path)

        results_json = []

        for result in results:
            for test_report in result.tests_results:
                status = ""
                if test_report.status == ResultStatus.PASS:
                    status = "pass"
                elif test_report.status == ResultStatus.BROK:
                    status = "brok"
                elif test_report.status == ResultStatus.WARN:
                    status = "warn"
                elif test_report.status == ResultStatus.CONF:
                    status = "conf"
                else:
                    status = "fail"

                data_test = {
                    "test_fqn": test_report.test.name,
                    "status": status,
                    "test": {
                        "command": test_report.test.command,
                        "arguments": test_report.test.arguments,
                        "log": test_report.stdout,
                        "retval": [str(test_report.return_code)],
                        "duration": test_report.exec_time,
                        "failed": test_report.failed,
                        "passed": test_report.passed,
                        "broken": test_report.broken,
                        "skipped": test_report.skipped,
                        "warnings": test_report.warnings,
                        "result": status,
                    },
                }

                results_json.append(data_test)

        data = {
            "results": results_json,
            "stats": {
                "runtime": sum(result.exec_time for result in results),
                "passed": sum(result.passed for result in results),
                "failed": sum(result.failed for result in results),
                "broken": sum(result.broken for result in results),
                "skipped": sum(result.skipped for result in results),
                "warnings": sum(result.warnings for result in results),
            },
            "environment": {
                "distribution": results[0].distro,
                "distribution_version": results[0].distro_ver,
                "kernel": results[0].kernel,
                "cmdline": results[0].cmdline,
                "arch": results[0].arch,
                "cpu": results[0].cpu,
                "swap": results[0].swap,
                "RAM": results[0].ram,
            },
        }

        async with AsyncFile(path, "w+") as outfile:
            text = json.dumps(data, indent=4)
            await outfile.write(text)

        self._logger.info("Report exported")



"""
.. module:: tempfile
    :platform: Linux
    :synopsis: module that contains temporary files handling

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import os
import pathlib
import pwd
import shutil
import tempfile
from typing import Optional


class TempDir:
    """
    Temporary directory handler.
    """

    SYMLINK_NAME = "latest"
    FOLDER_PREFIX = "kirk."

    def __init__(self, root: Optional[str], max_rotate: int = 5) -> None:
        """
        :param root: Root directory (i.e. /tmp). If None, TempDir will handle
            requests without adding any file or directory.
        :type root: str | None
        :param max_rotate: Maximum number of temporary directories.
        :type max_rotate: int
        """
        if root and not os.path.isdir(root):
            raise ValueError(f"root folder doesn't exist: {root}")

        self._root = os.path.abspath(root) if root else None
        self._max_rotate = max(max_rotate, 0)
        self._username = pwd.getpwuid(os.getuid()).pw_name if root else None
        self._folder = self._rotate()

    def _rotate(self) -> str:
        """
        Check for old folders and remove them, then create a new one and return
        its full path.
        """
        if not self._root:
            return ""

        tmpbase = os.path.join(self._root, f"{self.FOLDER_PREFIX}{self._username}")

        os.makedirs(tmpbase, exist_ok=True)

        # filter out symlink before sorting and counting
        all_paths = list(pathlib.Path(tmpbase).iterdir())
        sorted_paths = sorted(
            (p for p in all_paths if p.name != self.SYMLINK_NAME), key=os.path.getmtime
        )

        num_paths = len(sorted_paths)

        if num_paths >= self._max_rotate:
            max_items = num_paths - self._max_rotate + 1
            paths_to_remove = sorted_paths[:max_items]

            for path in paths_to_remove:
                shutil.rmtree(str(path.resolve()))

        # create a new folder
        folder = tempfile.mkdtemp(dir=tmpbase)

        # create symlink to the latest temporary directory
        latest = os.path.join(tmpbase, self.SYMLINK_NAME)
        if os.path.islink(latest):
            os.remove(latest)

        os.symlink(folder, latest, target_is_directory=True)

        return folder

    @property
    def root(self) -> str:
        """
        :return: The root folder. For example, if temporary folder is
            "/tmp/kirk.acer/tmpf547ftxv" the method will return "/tmp".
            If root folder has not been given during object creation, this
            method returns an empty string.
        :rtype: str
        """
        return self._root or ""

    @property
    def abspath(self) -> str:
        """
        :return: Absolute path of the temporary directory.
        :rtype: str
        """
        return self._folder

    def mkdir(self, path: str) -> None:
        """
        Create a directory inside temporary directory.

        :param path: Path of the directory.
        :type path: str
        :returns: Folder path.
        """
        if not self._folder:
            return

        os.mkdir(os.path.join(self._folder, path))

    def mkfile(self, path: str, content: str) -> None:
        """
        Create a file inside temporary directory.

        :param path: Path of the file.
        :type path: str
        :param content: File content.
        :type content: str
        """
        if not self._folder:
            return

        fpath = os.path.join(self._folder, path)
        with open(fpath, "w+", encoding="utf-8") as mypath:
            mypath.write(content)



"""
.. module:: sut_base
    :platform: Linux
    :synopsis: module implementing a generic SUT

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import (
    Any,
    Dict,
    Optional,
)

import libkirk.com
import libkirk.types
from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import SUTError
from libkirk.sut import SUT


class GenericSUT(SUT):
    """
    Generic SUT which is defining a structure to start implementing it.

    This class can be inherited in order to provide all its features also to
    the new implementations. Used by itself, it's just a SUT named 'default'
    for the kirk application and it's recognized by the plugin system.
    """

    def __init__(self) -> None:
        self._com = None

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        com_name = libkirk.types.dict_item(kwargs, "com", str, "shell")
        if not com_name:
            raise SUTError("Communication channel has not been defined")

        channels = libkirk.com.get_channels()
        if not channels:
            raise SUTError("No communication channels are provided")

        com = next((c for c in channels if c.name == com_name), None)
        if not com:
            raise SUTError(f"Can't find communication channel '{com_name}'")

        # pyrefly: ignore[bad-assignment]
        self._com = com

    @property
    def config_help(self) -> Dict[str, str]:
        return {
            "com": "Communication channel name (default: shell)",
        }

    @property
    def name(self) -> str:
        return "default"

    def get_channel(self) -> ComChannel:
        if not self._com:
            raise SUTError("SUT is not initialized")

        return self._com

    async def start(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if await self.is_running():
            return

        await self.get_channel().ensure_communicate(iobuffer)

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not await self.is_running():
            return

        await self.get_channel().stop(iobuffer)

    async def restart(self, iobuffer: Optional[IOBuffer] = None) -> None:
        await self.stop(iobuffer)
        await self.start(iobuffer)

    async def is_running(self) -> bool:
        return await self.get_channel().active()



"""
.. module:: events
    :platform: Linux
    :synopsis: events handler implementation module

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)


class Event:
    """
    An event to process.
    """

    def __init__(self, ordered: bool = False) -> None:
        """
        :param ordered: True if coroutines must be processed in order.
        :type ordered: bool
        """
        self._coros = []
        self._ordered = ordered

    def remove(self, coro: Callable) -> None:
        """
        Remove a specific Callable associated to the event.

        :param coro: Callable to remove.
        :type coro: Callable
        """
        try:
            self._coros.remove(coro)
        except ValueError:
            pass

    def has_coros(self) -> bool:
        """
        Check if there are still available registrations.

        :return: True if there are registered coroutines.
        :rtype: bool
        """
        return bool(self._coros)

    def register(self, coro: Callable) -> None:
        """
        Register a new Callable.

        :param coro: Coroutine to register.
        :type coro: Callable
        """
        self._coros.append(coro)

    def create_tasks(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> List[Any]:
        """
        Create tasks to run according to registered coroutines.

        :param args: Arguments to be passed to callback functions execution.
        :type args: list
        :param kwargs: Keyword arguments to be passed to callback functions
            execution.
        :type kwargs: dict
        :return: List of tasks to execute.
        :rtype: list(asyncio.Task)
        """
        if self._ordered:
            return [coro(*args, **kwargs) for coro in self._coros]

        tasks = [coro(*args, **kwargs) for coro in self._coros]
        return [asyncio.gather(*tasks)]


class EventsHandler:
    """
    This class implements event loop and events handling.
    """

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.events")
        self._tasks = asyncio.Queue()
        self._lock = asyncio.Lock()
        self._events: Dict[str, Event] = {}
        self._stop = False

        # register a default event used to notify internal
        # errors in the our application
        self._events["internal_error"] = Event()

    def _get_event(self, name: str) -> Optional[Event]:
        """
        Return an event according to its name.
        """
        return self._events.get(name, None)

    def reset(self) -> None:
        """
        Reset the entire events queue.
        """
        self._logger.info("Reset events queue")
        self._events.clear()

    def is_registered(self, event_name: str) -> bool:
        """
        Returns True if event_name is registered.

        :param event_name: Name of the event.
        :type event_name: str
        :return: True if registered, False otherwise.
        :rtype: bool
        """
        if not event_name:
            raise ValueError("event_name is empty")

        evt = self._get_event(event_name)
        return evt.has_coros() if evt else False

    def register(self, event_name: str, coro: Callable, ordered: bool = False) -> None:
        """
        Register an event with event_name.

        :param event_name: Name of the event.
        :type event_name: str
        :param coro: Callable associated with event_name.
        :type coro: Callable
        :param ordered: If True, the event will raise coroutines in the order
            they arrive.
        :type ordered: bool
        """
        if not event_name:
            raise ValueError("event_name is empty")

        if not coro:
            raise ValueError("coro is empty")

        self._logger.info("Register event: %s", repr(event_name))

        evt = self._events.setdefault(event_name, Event(ordered=ordered))
        evt.register(coro)

    def unregister(self, event_name: str, coro: Callable) -> None:
        """
        Unregister a single event Callable with event_name. If coro is None,
        all coroutines registered will be removed.

        :param event_name: Name of the event.
        :type event_name: str
        :param coro: Callable to unregister.
        :type coro: Callable
        """
        if not event_name:
            raise ValueError("event_name is empty")

        if not self.is_registered(event_name):
            raise ValueError(f"{event_name} is not registered")

        self._logger.info("Unregister event: %s -> %s", repr(event_name), repr(coro))

        if coro:
            self._events[event_name].remove(coro)
        else:
            del self._events[event_name]

    async def fire(self, event_name: str, *args: Any, **kwargs: Any) -> None:
        """
        Fire a specific event.

        :param event_name: Name of the event.
        :type event_name: str
        :param args: Arguments to be passed to callback functions execution.
        :type args: Any
        :param kwargs: Keyword arguments to be passed to callback functions
            execution.
        :type kwargs: Any
        """
        if not event_name:
            raise ValueError("event_name is empty")

        evt = self._get_event(event_name)
        if not evt:
            return

        for task in evt.create_tasks(*args, **kwargs):
            await self._tasks.put(task)

    async def _consume(self) -> None:
        """
        Consume the next event.
        """
        # asyncio.queue::get() will wait until an item is available
        # without blocking the application
        task = await self._tasks.get()
        if not task:
            return

        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception as err:
            if "internal_error" not in self._events:
                return

            self._logger.info("Exception catched")
            self._logger.error(err)

            ievt = self._get_event("internal_error")
            if ievt:
                tasks = ievt.create_tasks([err], task.get_name())
                await asyncio.gather(*tasks)
        finally:
            self._tasks.task_done()

    async def stop(self) -> None:
        """
        Stop the event loop.
        """
        self._logger.info("Stopping event loop")

        self._stop = True

        # indicate producer is done
        await self._tasks.put(None)

        async with self._lock:
            pass

        # consume the last tasks
        while not self._tasks.empty():
            await self._consume()

        self._logger.info("Event loop stopped")

    async def start(self) -> None:
        """
        Start the event loop.
        """
        self._stop = False

        try:
            async with self._lock:
                self._logger.info("Starting event loop")

                while not self._stop:
                    await self._consume()

                self._logger.info("Event loop completed")
        except asyncio.CancelledError:
            await self.stop()



"""
.. module:: data
    :platform: Linux
    :synopsis: module containing input data handling

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import (
    Dict,
    List,
    Optional,
)


class Test:
    """
    Test definition.
    """

    def __init__(
        self,
        name: str,
        cmd: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        args: Optional[List[str]] = None,
        parallelizable: bool = False,
    ) -> None:
        """
        :param name: Name of the test.
        :type name: str
        :param cmd: Command to execute.
        :type cmd: str
        :param cwd: Current working directory of the command.
        :type cwd: str
        :param env: Environment variables used to run the command.
        :type env: dict
        :param args: List of arguments.
        :type args: list(str)
        :param parallelizable: If True, test can be run in parallel.
        :type parallelizable: bool
        """
        if not name:
            raise ValueError("Test must have a name")

        if not cmd:
            raise ValueError("Test must have a command")

        self._name = name
        self._cmd = cmd
        self._cwd = cwd
        self._args = args if args else []
        self._env = env if env else {}
        self._parallelizable = parallelizable

    def __repr__(self) -> str:
        return (
            f"name: '{self._name}', "
            f"commmand: '{self._cmd}', "
            f"arguments: {self._args}, "
            f"cwd: '{self._cwd}', "
            f"environ: '{self._env}', "
            f"parallelizable: {self._parallelizable}"
        )

    @property
    def name(self) -> str:
        """
        :return: Name of the test.
        :rtype: str
        """
        return self._name

    @property
    def command(self) -> str:
        """
        :return: Command to execute test.
        :rtype: str
        """
        return self._cmd

    @property
    def arguments(self) -> List[str]:
        """
        :return: Arguments of the command.
        :rtype: list(str)
        """
        return self._args

    @property
    def parallelizable(self) -> bool:
        """
        :return: If True, test can be run in parallel.
        :rtype: bool
        """
        return self._parallelizable

    @property
    def cwd(self) -> Optional[str]:
        """
        :return: Current working directory.
        :rtype: str | None
        """
        return self._cwd

    @property
    def env(self) -> Dict[str, str]:
        """
        :return: Environment variables
        :rtype: dict
        """
        return self._env

    @property
    def full_command(self) -> str:
        """
        :return: Return the full command, with arguments as well.
            For example, if `command="ls"` and `arguments="-l -a"`,
            `full_command="ls -l -a"`.
        :rtype: str
        """
        cmd = self.command if self.command else ""
        if len(self.arguments) > 0:
            cmd += " "
            cmd += " ".join(self.arguments)

        return cmd

    def force_parallel(self) -> None:
        """
        :return: Force test to be parallelizable.
        """
        self._parallelizable = True


class Suite:
    """
    Testing suite definition class.
    """

    def __init__(self, name: str, tests: List[Test]) -> None:
        """
        :param name: Name of the testing suite.
        :type name: str
        :param tests: Tests of the suite.
        :type tests: list
        """
        self._name = name
        self._tests = tests

    def __repr__(self) -> str:
        return f"name: '{self._name}', tests: {self._tests}"

    @property
    def name(self) -> str:
        """
        :return: Name of the testing suite.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """
        Set the suite name.
        """
        if not value:
            raise ValueError("empty suite name")

        self._name = value

    @property
    def tests(self) -> List[Test]:
        """
        :return: Tests definitions.
        :rtype: list(Test)
        """
        return self._tests



"""
.. module:: ssh
    :platform: Linux
    :synopsis: module containing SSH channel implementation

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import contextlib
import importlib.util
import logging
import re
import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import libkirk.types
from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import (
    CommunicationError,
    KernelPanicError,
)

try:
    import asyncssh

    class MySSHClientSession(asyncssh.SSHClientSession):
        """
        Custom SSHClientSession used to store stdout during execution of
        commands and to check if Kernel Panic has occured in the system.
        """

        def __init__(self, iobuffer: Optional[IOBuffer] = None) -> None:
            self._output = []
            self._iobuffer = iobuffer
            self._panic = False
            self._logger = logging.getLogger("kirk.ssh.session")

        def data_received(self, data: str, datatype: asyncssh.DataType) -> None:
            """
            Override default data_received callback, storing stdout/stderr
            inside a buffer and checking for kernel panic.
            """
            self._output.append(data)

            if self._iobuffer:
                task = asyncio.create_task(self._iobuffer.write(data))
                task.add_done_callback(
                    lambda t: (
                        self._logger.error("IOBuffer write failed: %s", t.exception())
                        if not t.cancelled() and t.exception()
                        else None
                    )
                )

            if "Kernel panic" in data:
                self._panic = True

        def kernel_panic(self) -> bool:
            """
            True if command triggered a kernel panic during its execution.
            """
            return self._panic

        def get_output(self) -> List[str]:
            """
            Return the list containing stored stdout/stderr messages.
            """
            return self._output


except ModuleNotFoundError:
    pass


class SSHComChannel(ComChannel):
    """
    SSH communication channel.
    """

    _name = "ssh"

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.ssh")
        self._host = ""
        self._port = 22
        self._reset_cmd = ""
        self._user = ""
        self._password = ""
        self._key_file = ""
        self._sudo = False
        self._known_hosts: Optional[str] = None
        self._session_sem = asyncio.Semaphore()
        self._stop = False
        self._conn = None
        self._channels = []

    @property
    def config_help(self) -> Dict[str, str]:
        return {
            "host": "IP address of the host (default: localhost)",
            "port": "TCP port of the service (default: 22)",
            "user": "name of the user (default: root)",
            "password": "root password",
            "key_file": "private key location",
            "reset_cmd": "command to reset the remote target",
            "sudo": "use sudo to access to root shell (default: 0)",
            "known_hosts": "path to custom known_hosts file (optional)",
        }

    async def _reset(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Run the reset command on host.
        """
        if not self._reset_cmd:
            return

        self._logger.info("Executing reset command: %s", repr(self._reset_cmd))

        proc = await asyncio.create_subprocess_shell(
            self._reset_cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        if not proc or not proc.stdout:
            raise CommunicationError("Can't communicate with the host shell")

        while True:
            line = await proc.stdout.read(1024)
            if line:
                sline = line.decode(encoding="utf-8", errors="ignore")

                if iobuffer:
                    await iobuffer.write(sline)

            with contextlib.suppress(asyncio.TimeoutError):
                returncode = await asyncio.wait_for(proc.wait(), 1e-6)
                if returncode is not None:
                    break

        await proc.wait()

        self._logger.info("Reset command has been executed")

    def _create_command(
        self, cmd: str, cwd: Optional[str], env: Optional[Dict[str, Any]]
    ) -> str:
        """
        Create command to send to SSH client.
        """
        args = []

        if cwd:
            args.append(f"cd {cwd} && ")

        if env:
            for key, value in env.items():
                args.append(f"export {key}={value} && ")

        args.append(cmd)

        script = "".join(args)
        if self._sudo:
            script = f"sudo /bin/sh -c '{script}'"

        return script

    async def _read_max_sessions(self) -> int:
        """
        Read the SSH MaxSessions value and return it.
        """
        max_sessions = 10

        # pyrefly: ignore[missing-attribute]
        ret = await self._conn.run(
            "grep -i '^MaxSessions' /etc/ssh/sshd_config",
            timeout=5,
        )
        if ret.returncode == 0:
            match = re.search(r"^MaxSessions (?P<value>\d+)", ret.stdout)
            if match:
                max_sessions = int(match.group("value"))
                return max_sessions

        # pyrefly: ignore[missing-attribute]
        ret = await self._conn.run(
            "sudo sshd -T | grep maxsessions",
            timeout=5,
        )
        if ret.returncode == 0:
            match = re.search(r"^maxsessions (?P<value>\d+)", ret.stdout)
            if match:
                max_sessions = int(match.group("value"))

        self._logger.info("Maximum SSH sessions: %d", max_sessions)

        return max_sessions

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        if not importlib.util.find_spec("asyncssh"):
            raise CommunicationError("'asyncssh' library is not available")

        self._logger.info("Initialize SSH connection")

        self._host = libkirk.types.dict_item(kwargs, "host", str, default="localhost")
        self._reset_cmd = libkirk.types.dict_item(
            kwargs, "reset_cmd", str, default=None
        )
        self._user = libkirk.types.dict_item(kwargs, "user", str, default="root")
        self._password = libkirk.types.dict_item(kwargs, "password", str, default=None)
        self._key_file = libkirk.types.dict_item(kwargs, "key_file", str, default=None)
        self._known_hosts = libkirk.types.dict_item(
            kwargs, "known_hosts", str, default="~/.ssh/known_hosts"
        )

        if self._known_hosts == "/dev/null":
            self._known_hosts = None

        try:
            self._port = int(libkirk.types.dict_item(kwargs, "port", str, default="22"))

            if self._port < 1 or self._port > 65535:
                raise ValueError()
        except ValueError as err:
            raise CommunicationError(
                "'port' must be an integer between 1-65535"
            ) from err

        try:
            self._sudo = (
                int(libkirk.types.dict_item(kwargs, "sudo", str, default="0")) == 1
            )
        except ValueError as err:
            raise CommunicationError("'sudo' must be 0 or 1") from err

    @property
    def parallel_execution(self) -> bool:
        return True

    async def active(self) -> bool:
        return self._conn is not None

    async def communicate(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if await self.active():
            raise CommunicationError("SSH client is already connected")

        try:
            if self._key_file:
                priv_key = asyncssh.read_private_key(self._key_file)

                # pyrefly: ignore[bad-assignment]
                self._conn = await asyncssh.connect(
                    host=self._host,
                    port=self._port,
                    username=self._user,
                    client_keys=[priv_key],
                    known_hosts=self._known_hosts,
                )
            else:
                # pyrefly: ignore[bad-assignment]
                self._conn = await asyncssh.connect(
                    host=self._host,
                    port=self._port,
                    username=self._user,
                    password=self._password,
                    known_hosts=self._known_hosts,
                )

            max_sessions = await self._read_max_sessions()

            self._session_sem = asyncio.Semaphore(max_sessions)
        except (asyncssh.Error, ConnectionError) as err:
            if not self._stop:
                raise CommunicationError(err) from err

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not await self.active():
            return

        self._stop = True
        try:
            if self._channels:
                self._logger.info("Killing %d process(es)", len(self._channels))

                for proc in self._channels:
                    proc.kill()
                    await proc.wait_closed()

                self._channels.clear()

            self._logger.info("Closing connection")
            # pyrefly: ignore[missing-attribute]
            self._conn.close()
            # pyrefly: ignore[missing-attribute]
            await self._conn.wait_closed()
            self._logger.info("Connection closed")

            await self._reset(iobuffer=iobuffer)
        finally:
            self._stop = False
            self._conn = None

    async def ping(self) -> float:
        if not await self.active():
            raise CommunicationError("SSH client is not running")

        start_t = time.time()

        self._logger.info("Ping %s:%d", self._host, self._port)

        try:
            # pyrefly: ignore[missing-attribute]
            await self._conn.run("test .", check=True)
        except asyncssh.Error as err:
            raise CommunicationError(err) from err

        end_t = time.time() - start_t

        self._logger.info("Host replied after %.3f seconds", end_t)

        return end_t

    async def run_command(
        self,
        command: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        iobuffer: Optional[IOBuffer] = None,
    ) -> Optional[Dict[str, Any]]:
        if not command:
            raise ValueError("command is empty")

        if not await self.active():
            raise CommunicationError("SSH connection is not present")

        async with self._session_sem:
            cmd = self._create_command(command, cwd, env)
            start_t = 0
            stdout = []
            panic = False
            channel = None
            session = None
            ret = {
                "command": command,
                "returncode": 1,
                "exec_time": 0.0,
                "stdout": "",
            }

            try:
                self._logger.info("Running command: %s", repr(command))

                # pyrefly: ignore[missing-attribute]
                channel, session = await self._conn.create_session(
                    lambda: MySSHClientSession(iobuffer),
                    cmd,
                    encoding="utf-8",
                    errors="ignore",
                )

                self._channels.append(channel)
                start_t = time.time()

                await channel.wait_closed()

                panic = session.kernel_panic()
                stdout = session.get_output()
            except asyncssh.Error as err:
                if not self._stop:
                    raise CommunicationError(err)
            finally:
                if channel:
                    self._channels.remove(channel)
                    ret["returncode"] = channel.get_returncode()
                    ret["stdout"] = "".join(stdout)

                ret["exec_time"] = time.time() - start_t

            if panic:
                raise KernelPanicError()

            self._logger.info("Command executed")
            self._logger.debug(ret)

            return ret

    async def fetch_file(self, target_path: str) -> bytes:
        if not target_path:
            raise ValueError("target path is empty")

        if not await self.active():
            raise CommunicationError("SSH connection is not present")

        data = bytes()
        try:
            # pyrefly: ignore[missing-attribute]
            ret = await self._conn.run(f"cat {target_path}", check=True, encoding=None)

            data = bytes(ret.stdout)
        except asyncssh.Error as err:
            if not self._stop:
                raise CommunicationError(err) from err

        return data



"""
.. module:: monitor
    :platform: Linux
    :synopsis: modules used to generate real-time data from the executor

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import json
import logging
from typing import (
    Any,
    Dict,
)

import libkirk
from libkirk.data import (
    Suite,
    Test,
)
from libkirk.io import AsyncFile
from libkirk.results import (
    SuiteResults,
    TestResults,
)


class JSONFileMonitor:
    """
    Monitor the current executor status and it redirects events to a file
    using JSON format.
    """

    def __init__(self, path: str) -> None:
        """
        :param path: Path of the file.
        :type path: str
        """
        self._logging = logging.getLogger("libkirk.monitor")
        self._logging.info("File to monitor: %s", path)
        self._lock = asyncio.Lock()

        self._path = path
        self._events: Dict[str, Any] = {
            "session_restore": self.session_restore,
            "session_started": self.session_started,
            "session_stopped": self.session_stopped,
            "sut_stdout": self.sut_stdout,
            "sut_start": self.sut_start,
            "sut_stop": self.sut_stop,
            "sut_restart": self.sut_restart,
            "sut_not_responding": self.sut_not_responding,
            "run_cmd_start": self.run_cmd_start,
            "run_cmd_stop": self.run_cmd_stop,
            "test_stdout": self.test_stdout,
            "test_started": self.test_started,
            "test_completed": self.test_completed,
            "test_timed_out": self.test_timed_out,
            "suite_started": self.suite_started,
            "suite_completed": self.suite_completed,
            "suite_timeout": self.suite_timeout,
            "session_warning": self.session_warning,
            "session_error": self.session_error,
            "kernel_panic": self.kernel_panic,
            "kernel_tainted": self.kernel_tainted,
        }

    async def start(self) -> None:
        """
        Attach to events and start writing data inside the monitor file.
        """
        self._logging.info("Start monitoring")

        for name, coro in self._events.items():
            libkirk.events.register(name, coro)

    async def stop(self) -> None:
        """
        Stop monitoring events.
        """
        self._logging.info("Stop monitoring")

        for name, coro in self._events.items():
            libkirk.events.unregister(name, coro)

    async def _write(self, msg_type: str, msg: Dict[str, Any]) -> None:
        """
        Write a message to the JSON file.
        """
        data_str = json.dumps({"type": msg_type, "message": msg})

        async with self._lock:
            async with AsyncFile(self._path, "w") as fdata:
                await fdata.write(data_str)

    @staticmethod
    def _test_to_dict(test: Test) -> Dict[str, Any]:
        """
        Convert test into a dict which can be converted to JSON.
        """
        return {
            "name": test.name,
            "command": test.command,
            "arguments": test.arguments,
            "parallelizable": test.parallelizable,
            "cwd": test.cwd,
            "env": test.env,
        }

    def _suite_to_dict(self, suite: Suite) -> Dict[str, Any]:
        """
        Translate suite into a dict which can be converted into JSON.
        """
        return {
            "name": suite.name,
            "tests": [self._test_to_dict(test) for test in suite.tests],
        }

    async def session_restore(self, restore: str) -> None:
        await self._write("session_restore", {"restore": restore})

    async def session_started(self, suites: list, tmpdir: str) -> None:
        await self._write("session_started", {"tmpdir": tmpdir})

    async def session_stopped(self) -> None:
        await self._write("session_stopped", {})

    async def sut_stdout(self, sut: str, data: str) -> None:
        await self._write("sut_stdout", {"sut": sut, "data": data})

    async def sut_start(self, sut: str) -> None:
        await self._write("sut_start", {"sut": sut})

    async def sut_stop(self, sut: str) -> None:
        await self._write("sut_stop", {"sut": sut})

    async def sut_restart(self, sut: str) -> None:
        await self._write("sut_restart", {"sut": sut})

    async def sut_not_responding(self) -> None:
        await self._write("sut_not_responding", {})

    async def run_cmd_start(self, cmd: str) -> None:
        await self._write("run_cmd_start", {"cmd": cmd})

    async def run_cmd_stop(self, command: str, stdout: str, returncode: int) -> None:
        await self._write(
            "run_cmd_stop",
            {"command": command, "stdout": stdout, "returncode": returncode},
        )

    async def test_stdout(self, test: Test, data: str) -> None:
        await self._write(
            "test_stdout",
            {"test": self._test_to_dict(test), "data": data},
        )

    async def test_started(self, test: Test) -> None:
        await self._write("test_started", {"test": self._test_to_dict(test)})

    async def test_completed(self, results: TestResults) -> None:
        await self._write(
            "test_completed",
            {
                "test": self._test_to_dict(results.test),
                "stdout": results.stdout,
                "status": results.status,
                "exec_time": results.exec_time,
                "passed": results.passed,
                "failed": results.failed,
                "broken": results.broken,
                "skipped": results.skipped,
                "warnings": results.warnings,
            },
        )

    async def test_timed_out(self, test: Test, timeout: int) -> None:
        await self._write(
            "test_timed_out", {"test": self._test_to_dict(test), "timeout": timeout}
        )

    async def suite_started(self, suite: Suite) -> None:
        await self._write("suite_started", self._suite_to_dict(suite))

    async def suite_completed(self, results: SuiteResults, exec_time: float) -> None:
        # Optimize: Create dict directly in function call
        await self._write(
            "suite_completed",
            {
                "suite": self._suite_to_dict(results.suite),
                "exec_time": exec_time,
                "total_run": len(results.suite.tests),
                "passed": results.passed,
                "failed": results.failed,
                "skipped": results.skipped,
                "broken": results.broken,
                "warnings": results.warnings,
                "kernel_version": results.kernel,
                "cmdline": results.cmdline,
                "cpu": results.cpu,
                "arch": results.arch,
                "ram": results.ram,
                "swap": results.swap,
                "distro": results.distro,
                "distro_version": results.distro_ver,
            },
        )

    async def suite_timeout(self, suite: Suite, timeout: float) -> None:
        await self._write(
            "suite_timeout",
            {"suite": self._suite_to_dict(suite), "timeout": timeout},
        )

    async def session_warning(self, msg: str) -> None:
        await self._write("session_warning", {"message": msg})

    async def session_error(self, error: str) -> None:
        await self._write("session_error", {"error": error})

    async def kernel_panic(self) -> None:
        await self._write("kernel_panic", {})

    async def kernel_tainted(self, message: str) -> None:
        await self._write("kernel_tainted", {"message": message})



"""
.. module:: ui
    :platform: Linux
    :synopsis: module that contains user interface

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import platform
import sys
import traceback
from typing import (
    List,
    Optional,
)

import libkirk
from libkirk.data import (
    Suite,
    Test,
)
from libkirk.results import (
    SuiteResults,
    TestResults,
)


class ConsoleUserInterface:
    """
    Console based user interface.
    """

    WHITE = "\033[1;37m"
    GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    RED = "\033[1;31m"
    CYAN = "\033[1;36m"
    RESET_COLOR = "\033[0m"
    RESET_SCREEN = "\033[2J"

    def __init__(self, no_colors: bool = False) -> None:
        self._no_colors = no_colors
        self._line = ""
        self._restore = ""
        self._num_suites = 1

        event_handlers = {
            "session_restore": self.session_restore,
            "session_started": self.session_started,
            "session_stopped": self.session_stopped,
            "sut_start": self.sut_start,
            "sut_stop": self.sut_stop,
            "sut_restart": self.sut_restart,
            "run_cmd_start": self.run_cmd_start,
            "run_cmd_stdout": self.run_cmd_stdout,
            "run_cmd_stop": self.run_cmd_stop,
            "suite_started": self.suite_started,
            "suite_completed": self.suite_completed,
            "suite_timeout": self.suite_timeout,
            "session_warning": self.session_warning,
            "session_error": self.session_error,
            "session_completed": self.session_completed,
            "internal_error": self.internal_error,
        }

        for event_name, handler in event_handlers.items():
            libkirk.events.register(event_name, handler)

        # we register a special event 'printf' with ordered coroutines,
        # so we ensure that print threads will be executed one after the
        # other and user interface will be printed in the correct way
        libkirk.events.register("printf", self.print_message, ordered=True)

    async def _print(
        self, msg: str, color: Optional[str] = None, end: str = "\n"
    ) -> None:
        """
        Fire a `printf` event.
        """
        msg = msg.replace(self.RESET_SCREEN, "").replace("\r", "")

        if color and not self._no_colors:
            msg = f"{color}{msg}{self.RESET_COLOR}"

        await libkirk.events.fire("printf", msg, end=end, flush=True)

    async def print_message(
        self, msg: str, end: str = "\n", flush: bool = True
    ) -> None:
        """
        Print a message in console, avoiding any I/O blocking operation
        done by the `print` built-in function, using `asyncio.to_thread()`.
        """

        def _wrap() -> None:
            print(msg, end=end, flush=flush)

        await libkirk.to_thread(_wrap)

    @staticmethod
    def _user_friendly_duration(duration: float) -> str:
        """
        Return a user-friendly duration time from seconds.
        For example, "3670.234" becomes "1h 0m 10s".
        """
        if duration == 0:
            return "0h 0m 0s"

        minutes, seconds = divmod(duration, 60)
        hours, minutes = divmod(minutes, 60)

        if hours > 0:
            return f"{hours:.0f}h {minutes:.0f}m {seconds:.0f}s"
        elif minutes > 0:
            return f"{minutes:.0f}m {seconds:.0f}s"
        else:
            return f"{seconds:.3f}s"

    @staticmethod
    def _format_cmdline(cmdline: Optional[str]) -> str:
        """
        Format cmdline to show kernel parameters on multiple lines
        while preserving initial spaces.
        """
        if not cmdline:
            return ""

        parts = cmdline.split()
        if not parts:
            return cmdline

        formatted = parts[0]
        for part in parts[1:]:
            formatted += f"\n          {part}"

        return formatted

    def _result_color(self, results: TestResults) -> tuple:
        """
        Return test result string and the color associated to it.
        """
        if results.failed > 0:
            return "fail", self.RED
        if results.skipped > 0:
            return "skip", self.CYAN
        if results.broken > 0:
            return "broken", self.RED

        return "pass", self.GREEN

    async def _print_underline(self, msg: str) -> None:
        """
        Print an underlined message.
        """
        final_msg = f"{msg}\n{'─' * len(msg)}"
        await self._print(final_msg)

    async def _print_section(self, msg: str) -> None:
        """
        Print a section title surrounded by lines.
        """
        line = "─" * (len(msg) + 12)
        space = " " * 6
        await self._print(f"{line}\n{space}{msg}\n{line}")

    async def _print_target_info(self, results: SuiteResults) -> None:
        """
        Print target information.
        """
        message = (
            f"Kernel:   {results.kernel}\n"
            f"Cmdline:  {self._format_cmdline(results.cmdline)}\n"
            f"Machine:  {results.cpu}\n"
            f"Arch:     {results.arch}\n"
            f"RAM:      {results.ram}\n"
            f"Swap:     {results.swap}\n"
            f"Distro:   {results.distro} {results.distro_ver}\n"
        )

        await self._print_underline("Target information")
        await self._print(message)

    async def _print_summary(self, results: List[SuiteResults]) -> None:
        """
        Print a summary for a list of testing suites.
        """
        suites = ", ".join([s_res.suite.name for s_res in results])
        test_runs = sum(len(res.tests_results) for res in results)
        passed = sum(result.passed for result in results)
        failed = sum(result.failed for result in results)
        skipped = sum(result.skipped for result in results)
        broken = sum(result.broken for result in results)
        warnings = sum(result.warnings for result in results)
        exec_time = sum(result.exec_time for result in results)
        exec_time_uf = self._user_friendly_duration(exec_time)

        message = (
            f"Suite:   {suites}\n"
            f"Runtime: {exec_time_uf}\n"
            f"Runs:    {test_runs}\n\n"
            "Results:\n"
            f"    Passed:   {passed}\n"
            f"    Failed:   {failed}\n"
            f"    Broken:   {broken}\n"
            f"    Skipped:  {skipped}\n"
            f"    Warnings: {warnings}\n"
        )
        await self._print(message)

    async def session_restore(self, restore: str) -> None:
        await self._print(f"Restore session: {restore}")

    async def session_started(self, suites: list, tmpdir: str) -> None:
        self._num_suites = len(suites) if suites is not None else 0

        uname = platform.uname()
        message = (
            "Host information\n"
            f"\tHostname:   {uname.node}\n"
            f"\tPython:     {sys.version}\n"
            f"\tDirectory:  {tmpdir}\n"
        )

        await self._print(message)

    async def session_stopped(self) -> None:
        await self._print("Session stopped")

    async def sut_start(self, sut: str) -> None:
        await self._print(f"Connecting to SUT: {sut}\n")

    async def sut_stop(self, sut: str) -> None:
        await self._print(f"Disconnecting from SUT: {sut}")

    async def sut_restart(self, sut: str) -> None:
        await self._print(f"Restarting SUT: {sut}")

    async def run_cmd_start(self, cmd: str) -> None:
        await self._print(f"{cmd}", color=self.CYAN)

    async def run_cmd_stdout(self, data: str) -> None:
        await self._print(data, end="")

    async def run_cmd_stop(self, command: str, stdout: str, returncode: int) -> None:
        await self._print(f"\nExit code: {returncode}\n")

    async def suite_started(self, suite: Suite) -> None:
        suite_msg = f"Suite: {suite.name}"
        await self._print_underline(suite_msg)

    async def suite_completed(self, results: SuiteResults, exec_time: float) -> None:
        exec_time_uf = self._user_friendly_duration(exec_time)

        message = f"\nExecution time: {exec_time_uf}\n"

        await self._print(message)

        # there's no need to print more than one summary if we only have
        # one testing suite
        if self._num_suites > 1:
            await self._print_summary([results])

    async def suite_timeout(self, suite: Suite, timeout: float) -> None:
        await self._print(
            f"Suite '{suite.name}' timed out after {timeout} seconds", color=self.RED
        )

    async def session_warning(self, msg: str) -> None:
        await self._print(f"Warning: {msg}", color=self.YELLOW)

    async def session_error(self, error: str) -> None:
        await self._print(f"Error: {error}", color=self.RED)

    async def session_completed(self, results: List[SuiteResults]) -> None:
        if not results:
            return

        await self._print("")
        await self._print_target_info(results[0])
        await self._print_section("TEST SUMMARY")
        await self._print_summary(results)

        t_broken = []
        t_failed = []

        for s_res in results:
            for t_res in s_res.tests_results:
                if t_res.failed > 0:
                    t_failed.append(t_res)
                if t_res.broken > 0:
                    t_broken.append(t_res)

        if t_broken:
            await self._print("Broken:", color=self.RED)
            msg = [f"    • {t.test.name}" for t in t_broken]
            await self._print("\n".join(msg))
            await self._print("")

        if t_failed:
            await self._print("Failures:", color=self.RED)
            msg = [f"    • {t.test.name}" for t in t_failed]
            await self._print("\n".join(msg))
            await self._print("")

    async def internal_error(self, exc: BaseException, func_name: str) -> None:
        await self._print(
            f"\nUI error in function '{func_name}': {exc}\n", color=self.RED
        )

        traceback.print_exc()


class SimpleUserInterface(ConsoleUserInterface):
    """
    Console based user interface without many fancy stuff.
    """

    def __init__(self, no_colors: bool = False) -> None:
        super().__init__(no_colors=no_colors)

        self._sut_not_responding = False
        self._kernel_panic = False
        self._kernel_tainted: Optional[str] = None
        self._timed_out = False

        libkirk.events.register("sut_not_responding", self.sut_not_responding)
        libkirk.events.register("kernel_panic", self.kernel_panic)
        libkirk.events.register("kernel_tainted", self.kernel_tainted)
        libkirk.events.register("test_timed_out", self.test_timed_out)
        libkirk.events.register("test_started", self.test_started)
        libkirk.events.register("test_completed", self.test_completed)

    async def sut_not_responding(self) -> None:
        self._sut_not_responding = True
        # this message will replace ok/fail message
        await self._print("SUT not responding", color=self.RED)

    async def kernel_panic(self) -> None:
        self._kernel_panic = True
        # this message will replace ok/fail message
        await self._print("kernel panic", color=self.RED)

    async def kernel_tainted(self, message: str) -> None:
        self._kernel_tainted = message

    async def test_timed_out(self, _: Test, timeout: int) -> None:
        self._timed_out = True
        # this message will replace ok/fail message
        await self._print("timed out", color=self.RED)

    async def test_started(self, test: Test) -> None:
        await self._print(f"{test.name}: ", color=self.WHITE, end="")

    async def test_completed(self, results: TestResults) -> None:
        if self._timed_out or self._sut_not_responding or self._kernel_panic:
            self._sut_not_responding = False
            self._kernel_panic = False
            self._timed_out = False
            return

        msg, col = self._result_color(results)
        await self._print(msg, color=col, end="")

        if self._kernel_tainted:
            await self._print(" | ", end="")
            await self._print("tainted", color=self.YELLOW, end="")
            self._kernel_tainted = None

        uf_time = self._user_friendly_duration(results.exec_time)
        await self._print(f"  ({uf_time})")


class VerboseUserInterface(ConsoleUserInterface):
    """
    Verbose console based user interface.
    """

    def __init__(self, no_colors: bool = False) -> None:
        super().__init__(no_colors=no_colors)

        self._timed_out = False

        libkirk.events.register("sut_stdout", self.sut_stdout)
        libkirk.events.register("kernel_tainted", self.kernel_tainted)
        libkirk.events.register("test_timed_out", self.test_timed_out)
        libkirk.events.register("test_started", self.test_started)
        libkirk.events.register("test_completed", self.test_completed)
        libkirk.events.register("test_stdout", self.test_stdout)

    async def sut_stdout(self, _: str, data: str) -> None:
        await self._print(data, end="")

    async def kernel_tainted(self, message: str) -> None:
        await self._print(f"Tainted kernel: {message}", color=self.YELLOW)

    async def test_timed_out(self, _: Test, timeout: int) -> None:
        self._timed_out = True

    async def test_started(self, test: Test) -> None:
        await self._print_section(test.name)
        await self._print("Executing: ", end="")
        await self._print(test.full_command, end="\n\n")

    async def test_completed(self, results: TestResults) -> None:
        if self._timed_out:
            await self._print("Test timed out", color=self.RED)

        self._timed_out = False

        parts = []

        if "Summary:" not in results.stdout:
            parts.extend(
                [
                    "\nSummary:",
                    f"passed    {results.passed}",
                    f"failed    {results.failed}",
                    f"broken    {results.broken}",
                    f"skipped   {results.skipped}",
                    f"warnings  {results.warnings}",
                ]
            )

        uf_time = self._user_friendly_duration(results.exec_time)
        parts.append(f"\nDuration: {uf_time}\n")

        await self._print("\n".join(parts))

    async def test_stdout(self, _: Test, data: str) -> None:
        await self._print(data, end="")


class ParallelUserInterface(ConsoleUserInterface):
    """
    Console based user interface for parallel execution of the tests.
    """

    def __init__(self, no_colors: bool = False) -> None:
        super().__init__(no_colors=no_colors)

        self._sut_not_responding = False
        self._kernel_panic = False
        self._kernel_tainted: Optional[str] = None
        self._timed_out = False
        self._pl_total = 0
        self._pl_done = 0

        libkirk.events.register("sut_not_responding", self.sut_not_responding)
        libkirk.events.register("kernel_panic", self.kernel_panic)
        libkirk.events.register("kernel_tainted", self.kernel_tainted)
        libkirk.events.register("suite_started", self.print_parallel)
        libkirk.events.register("test_timed_out", self.test_timed_out)
        libkirk.events.register("test_completed", self.test_completed)

    async def sut_not_responding(self) -> None:
        self._sut_not_responding = True

    async def kernel_panic(self) -> None:
        self._kernel_panic = True

    async def kernel_tainted(self, message: str) -> None:
        self._kernel_tainted = message

    async def test_timed_out(self, _: Test, timeout: int) -> None:
        self._timed_out = True

    async def print_parallel(self, suite: Suite) -> None:
        parallel_tests = [
            f"• {test.name}" for test in suite.tests if test.parallelizable
        ]

        if parallel_tests:
            self._pl_total += len(parallel_tests)
            await self._print("Following tests will run in parallel:")
            await self._print("\n".join(parallel_tests), end="\n\n")

    async def test_completed(self, results: TestResults) -> None:
        if results.test.parallelizable:
            self._pl_done += 1

            await self._print(
                f"{results.test.name} ({self._pl_done}/{self._pl_total}): ", end=""
            )
        else:
            await self._print(f"{results.test.name}: ", end="")

        if self._timed_out:
            await self._print("timed out", color=self.RED)
        elif self._sut_not_responding:
            # this message will replace ok/fail message
            await self._print("SUT not responding", color=self.RED)
        elif self._kernel_panic:
            # this message will replace ok/fail message
            await self._print("kernel panic", color=self.RED)
        else:
            msg, col = self._result_color(results)
            if self._kernel_tainted:
                await self._print(" | ", end="")
                await self._print("tainted", color=self.YELLOW, end="")

            uf_time = self._user_friendly_duration(results.exec_time)
            await self._print(f"  ({uf_time})")

        self._sut_not_responding = False
        self._kernel_panic = False
        self._kernel_tainted = None
        self._timed_out = False



"""
.. module:: sut
    :platform: Linux
    :synopsis: module implementing SUT

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import re
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

import libkirk.plugin
from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import SUTError
from libkirk.plugin import Plugin

# discovered SUT implementations
_SUT = []


class SUT(Plugin):
    """
    SUT abstraction class. It could be a remote host, a local host, a virtual
    machine instance, or any complex system we want to test.
    """

    TAINTED_MSG = [
        "proprietary module was loaded",
        "module was force loaded",
        "kernel running on an out of specification system",
        "module was force unloaded",
        "processor reported a Machine Check Exception (MCE)",
        "bad page referenced or some unexpected page flags",
        "taint requested by userspace application",
        "kernel died recently, i.e. there was an OOPS or BUG",
        "ACPI table overridden by user",
        "kernel issued warning",
        "staging driver was loaded",
        "workaround for bug in platform firmware applied",
        "externally-built (“out-of-tree”) module was loaded",
        "unsigned module was loaded",
        "soft lockup occurred",
        "kernel has been live patched",
        "auxiliary taint, defined for and used by distros",
        "kernel was built with the struct randomization plugin",
    ]

    FAULT_INJECTION_FILES = [
        "fail_io_timeout",
        "fail_make_request",
        "fail_page_alloc",
        "failslab",
    ]

    _optimize = False

    def get_channel(self) -> ComChannel:
        """
        :return: Main channel to communicated with SUT.
        :rtype: ComChannel
        """
        raise NotImplementedError()

    async def start(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Start the SUT.

        :param iobuffer: IO channel where to write stdout.
        :type iobuffer: IOBuffer
        """
        raise NotImplementedError()

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Stop the SUT.

        :param iobuffer: IO channel where to write stdout.
        :type iobuffer: IOBuffer
        """
        raise NotImplementedError()

    async def restart(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Restart the SUT.

        :param iobuffer: IO channel where to write stdout.
        :type iobuffer: IOBuffer
        """
        raise NotImplementedError()

    async def is_running(self) -> bool:
        """
        :return: True if system under test is up and running. False otherwise.
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    def optimize(self) -> bool:
        """
        Optimize the commands execution by applying parallelization
        when it's available.

        :return: True if SUT will optimize commands execution on SUT.
        """
        return self._optimize

    @optimize.setter
    def optimize(self, value: bool) -> None:
        """
        Set commands execution optimization.
        """
        self._optimize = value

    async def _run_cmd(self, cmd: str) -> str:
        """
        Run command, check for returncode and return command's stdout.
        """
        stdout = "unknown"
        try:
            channel = self.get_channel()
            ret = await asyncio.wait_for(channel.run_command(cmd), 1.5)
            if ret and ret["returncode"] == 0:
                stdout = ret["stdout"].rstrip()
        except asyncio.TimeoutError:
            pass

        return stdout

    async def _get_distro(self) -> str:
        """
        Return the distro name.
        """
        return await self._run_cmd('. /etc/os-release && echo "$ID"')

    async def _get_distro_ver(self) -> str:
        """
        Return the distro version.
        """
        return await self._run_cmd('. /etc/os-release && echo "$VERSION_ID"')

    async def _get_kernel(self) -> str:
        """
        Return the kernel name.
        """
        return await self._run_cmd("uname -s -r -v")

    async def _get_cmdline(self) -> str:
        """
        Return /proc/cmdline content.
        """
        return await self._run_cmd("cat /proc/cmdline")

    async def _get_arch(self) -> str:
        """
        Return the architecture name.
        """
        return await self._run_cmd("uname -m")

    async def _get_cpu(self) -> str:
        """
        Return the CPU name.
        """
        return await self._run_cmd("uname -p")

    async def _get_meminfo(self) -> str:
        """
        Return the memory information.
        """
        return await self._run_cmd("cat /proc/meminfo")

    async def get_info(self) -> Dict[str, str]:
        """
        Return SUT information.

        :return: Dictionary containing the SUT information in the form of:

            .. code-block:: python

                {
                    "distro": str,
                    "distro_ver": str,
                    "kernel": str,
                    "cmdline": str,
                    "arch": str,
                    "cpu" : str,
                    "swap" : str,
                    "ram" : str,
                }

        :rtype: dict
        """
        if not await self.is_running():
            raise SUTError("SUT is not running")

        distro = ""
        distro_ver = ""
        kernel = ""
        cmdline = ""
        arch = ""
        cpu = ""
        meminfo = ""

        if self.optimize:
            # pyrefly: ignore[bad-unpacking]
            (
                distro,
                distro_ver,
                kernel,
                cmdline,
                arch,
                cpu,
                meminfo,
            ) = await asyncio.gather(
                *[
                    self._get_distro(),
                    self._get_distro_ver(),
                    self._get_kernel(),
                    self._get_cmdline(),
                    self._get_arch(),
                    self._get_cpu(),
                    self._get_meminfo(),
                ]
            )
        else:
            distro = await self._get_distro()
            distro_ver = await self._get_distro_ver()
            kernel = await self._get_kernel()
            cmdline = await self._get_cmdline()
            arch = await self._get_arch()
            cpu = await self._get_cpu()
            meminfo = await self._get_meminfo()

        memory = "unknown"
        swap = "unknown"

        if meminfo:
            mem_m = re.search(r"MemTotal:\s+(?P<memory>\d+\s+kB)", meminfo)
            if mem_m:
                memory = mem_m.group("memory")

            swap_m = re.search(r"SwapTotal:\s+(?P<swap>\d+\s+kB)", meminfo)
            if swap_m:
                swap = swap_m.group("swap")

        ret = {
            "distro": distro,
            "distro_ver": distro_ver,
            "kernel": kernel,
            "cmdline": cmdline,
            "arch": arch,
            "cpu": cpu,
            "ram": memory,
            "swap": swap,
        }

        return ret

    # Lazily initialised inside get_tainted_info() so that the
    # asyncio primitives are always created on the *running* loop.
    # Class-level assignment would bind them to the loop that was
    # current at import time, which breaks on Python 3.7-3.9 when
    # tests run on a different loop than the session loop.
    _tainted_lock: Optional[asyncio.Lock] = None
    _tainted_status: Optional[asyncio.Queue] = None

    async def get_tainted_info(self) -> Tuple[int, List[str]]:
        """
        Return information about kernel if tainted.

        :return: A tuple containing tainted information. First element is the
            tainted code and the second element is the tainted message.
        :rtype: (int, list[str])
        """
        if not await self.is_running():
            raise SUTError("SUT is not running")

        # Initialise on first use, always inside a running loop.
        if self._tainted_lock is None:
            self._tainted_lock = asyncio.Lock()
        if self._tainted_status is None:
            self._tainted_status = asyncio.Queue(maxsize=1)

        if self._tainted_lock.locked():
            try:
                status = self._tainted_status.get_nowait()
                return status
            except asyncio.QueueEmpty:
                pass

        async with self._tainted_lock:
            channel = self.get_channel()
            ret = await channel.run_command("cat /proc/sys/kernel/tainted")
            if not ret or ret["returncode"] != 0:
                raise SUTError("Can't read tainted kernel information")

            tainted_num = len(self.TAINTED_MSG)
            code = ret["stdout"].strip()

            # output is likely message in stderr
            if not code.isdigit():
                raise SUTError(code)

            code = int(code)
            bits = format(code, f"0{tainted_num}b")[::-1]

            messages = []
            for i in range(0, tainted_num):
                if bits[i] == "1":
                    msg = self.TAINTED_MSG[i]
                    messages.append(msg)

            try:
                self._tainted_status.get_nowait()
            except asyncio.QueueEmpty:
                pass

            await self._tainted_status.put((code, messages))

            return code, messages

    async def logged_as_root(self) -> bool:
        """
        :return: True if we are logged as root inside the SUT. False otherwise.
        :rtype: bool
        """
        if not await self.is_running():
            raise SUTError("SUT is not running")

        channel = self.get_channel()
        ret = await channel.run_command("id -u")
        if not ret or ret["returncode"] != 0:
            raise SUTError("Can't determine if we are running as root")

        val = ret["stdout"].rstrip()
        user_id = 100
        try:
            user_id = int(val)
        except ValueError as err:
            raise SUTError(f"'id -u' returned {val}") from err

        return user_id == 0

    async def is_fault_injection_enabled(self) -> bool:
        """
        :return: True if fault injection is enabled in the kernel. False
            otherwise.
        :rtype: bool
        """
        if not await self.is_running():
            raise SUTError("SUT is not running")

        channel = self.get_channel()

        for ftype in self.FAULT_INJECTION_FILES:
            ret = await channel.run_command(f"test -d /sys/kernel/debug/{ftype}")
            if ret and ret["returncode"] != 0:
                return False

        return True

    async def setup_fault_injection(self, prob: int) -> None:
        """
        Configure kernel fault injection. When prob is zero, the fault
        injection is set to default values.

        :param prob: Fault probabilty in between 0-100.
        :type prob: int
        """
        if not await self.is_running():
            raise SUTError("SUT is not running")

        interval = 1 if prob == 0 else 100
        times = 1 if prob == 0 else -1

        async def _set_value(value: int, path: str) -> None:
            """
            Set the value to the path
            """
            channel = self.get_channel()
            ret = await channel.run_command(f"echo {value} > {path}")
            if ret and ret["returncode"] != 0:
                raise SUTError(f"Can't setup {path}. {ret['stdout']}")

        for ftype in self.FAULT_INJECTION_FILES:
            path = f"/sys/kernel/debug/{ftype}"

            await _set_value(0, f"{path}/space")
            await _set_value(times, f"{path}/times")
            await _set_value(interval, f"{path}/interval")
            await _set_value(prob, f"{path}/probability")


def discover(path: str, extend: bool = True) -> None:
    """
    Discover all SUT implementations inside path.

    :param path: Directory where searching for SUT implementations.
    :type path: str
    :param extend: If True, it will add new discovered SUT on top of the
        ones already found. If False, previous discovered SUT will be
        cleared.
    :type param: bool
    """
    global _SUT

    obj = libkirk.plugin.discover(SUT, path)
    if not extend:
        _SUT.clear()

    _SUT.extend(obj)


def get_suts() -> List[SUT]:
    """
    :return: List of loaded SUT implementations.
    :rtype: list(SUT)
    """
    global _SUT
    return _SUT



"""
.. module:: com
    :platform: Linux
    :synopsis: communication class definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import libkirk.plugin
from libkirk.errors import (
    KirkException,
    PluginError,
)
from libkirk.plugin import Plugin

# discovered communication channels
_COM = []


class IOBuffer:
    """
    IO stdout buffer. The API is similar to IO types.
    """

    async def write(self, data: str) -> None:
        """
        Write data inside the buffer.

        :param data: Data to write.
        :type data: str
        """
        raise NotImplementedError()


class ComChannel(Plugin):
    """
    Communication channel. The objects implementing this class are usually
    using SSH, serial, shell, etc protocols. and they are used by the scheduler
    in order to execute commands or turning on/off the communication.
    """

    @property
    def parallel_execution(self) -> bool:
        """
        :return: If True, communication supports commands parallel execution.
        :rtype: bool
        """
        raise NotImplementedError()

    async def active(self) -> bool:
        """
        :return: Return True if communication is active. False otherwise.
        :rtype: bool
        """
        raise NotImplementedError()

    async def communicate(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Start communication.

        :param iobuffer: Buffer used to write stdout.
        :type iobuffer: IOBuffer
        """
        raise NotImplementedError()

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        """
        Stop communication.

        :param iobuffer: Buffer used to write stdout.
        :type iobuffer: IOBuffer
        """
        raise NotImplementedError()

    async def ping(self) -> float:
        """
        Send a ping request and verify how much reply takes in seconds.

        :return: Time between ping and pong.
        :rtype: float
        """
        raise NotImplementedError()

    async def run_command(
        self,
        command: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        iobuffer: Optional[IOBuffer] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Run a command.

        :param command: Command to execute.
        :type command: str
        :param cwd: Current working directory.
        :type cwd: str
        :param env: Environment variables.
        :type env: dict
        :param iobuffer: Buffer used to write stdout.
        :type iobuffer: IOBuffer
        :return: Dictionary containing information about the executed command.

            .. code-block:: python

                {
                    "command": <str>,
                    "returncode": <int>,
                    "stdout": <str>,
                    "exec_time": <float>,
                }

            If None is returned, then callback has failed.
        :rtype: dict
        """
        raise NotImplementedError()

    async def fetch_file(self, target_path: str) -> bytes:
        """
        Fetch file and return its content.

        :param target_path: Path of the file to download from target.
        :type target_path: str
        :return: Data contained in target_path.
        :rtype: bytes
        """
        raise NotImplementedError()

    async def ensure_communicate(
        self, iobuffer: Optional[IOBuffer] = None, retries: int = 10
    ) -> None:
        """
        Ensure that communicate is completed, retrying as many times we
        want in case of KirkException error. After each error, the
        communication is stopped and a new communication is performed.

        :param iobuffer: Buffer used to write stdout.
        :type iobuffer: IOBuffer
        :param retries: Number of times we retry to communicate.
        :type retries: int
        """
        retries = max(retries, 1)

        for retry in range(retries):
            try:
                await self.communicate(iobuffer=iobuffer)
                break
            except KirkException as err:
                if retry >= retries - 1:
                    raise err

                await self.stop(iobuffer=iobuffer)


def discover(path: str, extend: bool = True) -> None:
    """
    Discover all ComChannel implementations inside `path`.

    :param path: Directory where searching for channel implementations.
    :type path: str
    :param extend: If True, it will add new discovered channels on top of the
        ones already found. If False, previous discovered channels will be
        cleared.
    :rtype: bool
    """
    global _COM

    obj = libkirk.plugin.discover(ComChannel, path)
    if not extend:
        _COM.clear()

    _COM.extend(obj)


def get_channels() -> List[ComChannel]:
    """
    :return: List of loaded ComChannel implementations.
    :rtype: list(ComChannel)
    """
    global _COM
    # pyrefly: ignore[bad-return]
    return _COM


def clone_channel(name: str, new_name: str) -> Plugin:
    """
    Clone a channel implementation named name and rename it with
    new_name. The new plugin will be registered with the other
    plugins.

    :param name: Plugin name.
    :type name: str
    :param new_name: New cloned plugin name.
    :type new_name: str
    :return: New plugin object.
    :rtype: Plugin
    """
    global _COM

    plugin = next((c for c in _COM if c.name == name), None)
    if not plugin:
        raise PluginError(f"Can't find plugin '{name}'")

    channel = plugin.clone(new_name)
    _COM.append(channel)

    return channel



"""
.. module:: session
    :platform: Linux
    :synopsis: session declaration

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import copy
import logging
import os
import random
import re
from typing import (
    Dict,
    List,
    Optional,
)

import libkirk
import libkirk.types
from libkirk.data import Suite
from libkirk.errors import (
    KirkException,
    SessionError,
)
from libkirk.export import JSONExporter
from libkirk.io import AsyncFile
from libkirk.ltp import LTPFramework
from libkirk.results import TestResults
from libkirk.scheduler import SuiteScheduler
from libkirk.sut import (
    SUT,
    IOBuffer,
)
from libkirk.tempfile import TempDir


class RedirectSUTStdout(IOBuffer):
    """
    Redirect stdout data to UI events.
    """

    def __init__(self, sut: SUT, is_cmd: bool) -> None:
        self._sut = sut
        self._is_cmd = is_cmd

    async def write(self, data: str) -> None:
        if self._is_cmd:
            await libkirk.events.fire("run_cmd_stdout", data)
        else:
            await libkirk.events.fire("sut_stdout", self._sut.name, data)


class Session:
    """
    The session runner.
    """

    def __init__(
        self,
        tmpdir: TempDir,
        sut: SUT,
        exec_timeout: float = 3600.0,
        suite_timeout: float = 3600.0,
        workers: int = 1,
        force_parallel: bool = False,
    ) -> None:
        """
        :param tmpdir: Temporary directory.
        :type tmpdir: TempDir
        :param sut: SUT communication object.
        :type sut: SUT
        :param exec_timeout: Test timeout.
        :type exec_timeout: float
        :param suite_timeout: Testing suite timeout.
        :type suite_timeout: float
        :param workers: Number of workers for testing suite scheduler.
        :type workers: int
        :param force_parallel: Force parallel execution of all tests.
        :type force_parallel: bool
        """
        self._logger = logging.getLogger("kirk.session")
        self._tmpdir = tmpdir
        self._sut = sut
        self._exec_timeout = exec_timeout
        self._force_parallel = force_parallel
        self._stop = False
        self._exec_lock = asyncio.Lock()
        self._run_lock = asyncio.Lock()
        self._results = []
        self._framework = LTPFramework(
            timeout=self._exec_timeout,
        )

        self._scheduler = SuiteScheduler(
            sut=self._sut,
            framework=self._framework,
            suite_timeout=suite_timeout,
            exec_timeout=self._exec_timeout,
            max_workers=workers,
        )

        self._curr_suite = ""
        self._setup_debug_log()
        self._setup_test_save()

        if not self._sut.get_channel().parallel_execution:
            self._logger.info(
                "SUT doesn't support parallel execution. Forcing workers=1."
            )

    def _setup_debug_log(self) -> None:
        """
        Set logging module so we save a log file with debugging information
        inside the temporary path.
        """
        if not self._tmpdir.abspath:
            return

        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        debug_file = os.path.join(self._tmpdir.abspath, "debug.log")
        handler = logging.FileHandler(debug_file, encoding="utf8")
        handler.setLevel(logging.DEBUG)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s:%(lineno)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    def _setup_test_save(self) -> None:
        """
        Setup event for complete test saving.
        """
        if not self._tmpdir.abspath:
            return

        async def save_suite_started(suite: Suite) -> None:
            self._curr_suite = suite.name

        async def save_test_file(results: TestResults) -> None:
            epath = os.path.join(self._tmpdir.abspath, "executed")
            async with AsyncFile(epath, "a+") as efile:
                await efile.write(f"{self._curr_suite}::{results.test.name}\n")

        libkirk.events.register("suite_started", save_suite_started)
        libkirk.events.register("test_completed", save_test_file)

    async def _read_restored_session(self, path: str) -> Dict[str, List[str]]:
        """
        Read restored session.
        """
        data = {}
        if not (path and os.path.exists(path)):
            return data

        epath = os.path.join(path, "executed")
        if not os.path.exists(epath):
            return data

        self._logger.info("Reading previous executed tests")

        async with AsyncFile(epath, "r") as efile:
            async for line in efile:
                if not line or "::" not in line:
                    continue

                suite, test = line.split("::")
                if not (suite and test):
                    continue

                if suite not in data:
                    data[suite] = []

                data[suite].append(test.rstrip())

        self._logger.debug(data)

        return data

    async def _start_sut(self) -> None:
        """
        Start communicating with SUT.
        """
        await libkirk.events.fire("sut_start", self._sut.name)
        await self._sut.start(iobuffer=RedirectSUTStdout(self._sut, False))

    async def _stop_sut(self) -> None:
        """
        Stop the SUT.
        """
        if not await self._sut.is_running():
            return

        await libkirk.events.fire("sut_stop", self._sut.name)
        await self._sut.stop(iobuffer=RedirectSUTStdout(self._sut, False))

    async def _get_suites_objects(self, names: List[str]) -> List[Suite]:
        """
        Return suites objects by giving their names.
        """
        channel = self._sut.get_channel()
        coros = [self._framework.find_suite(channel, suite) for suite in names]

        if not coros:
            raise KirkException(f"Can't find suites: {names}")

        results = await asyncio.gather(*coros, return_exceptions=True)
        for suite in results:
            if isinstance(suite, Exception):
                raise suite

            if not suite:
                raise KirkException("Can't find suite objects")

        # pyrefly: ignore=no-matching-overload
        return list(results)

    async def _restore_tests(
        self, suites_obj: list, restore_path: Optional[str] = None
    ) -> None:
        """
        Remove all tests but the one which need to be restored.
        """
        if not restore_path:
            return

        restored = await self._read_restored_session(restore_path)
        if not restored:
            return

        await libkirk.events.fire("session_restore", restore_path)

        for suite_obj in suites_obj:
            suite = suite_obj.name
            if suite not in restored:
                continue

            restored_set = set(restored[suite])
            suite_obj.tests[:] = [
                test for test in suite_obj.tests if test.name not in restored_set
            ]

    @staticmethod
    def _filter_tests(
        suites_obj: List[Suite],
        regex: Optional[str] = None,
        when_matching: bool = False,
    ) -> None:
        """
        Filter tests according to `regex`, if `when_matching` is True.
        """
        if not regex:
            return

        matcher = re.compile(regex)

        for suite_obj in suites_obj:
            if when_matching:
                # skip matching tests (keep non-matching)
                suite_obj.tests[:] = [
                    test for test in suite_obj.tests if not matcher.search(test.name)
                ]
            else:
                # keep only matching tests
                suite_obj.tests[:] = [
                    test for test in suite_obj.tests if matcher.search(test.name)
                ]

    @staticmethod
    def _apply_iterate(suites_obj: List[Suite], suite_iterate: int) -> List[Suite]:
        """
        Return testing suites after applying iterate parameters.
        """
        if suite_iterate <= 1:
            return suites_obj

        suites_list = []
        for suite in suites_obj:
            for i in range(suite_iterate):
                obj = copy.deepcopy(suite)
                obj.name = f"{suite.name}[{i}]"
                suites_list.append(obj)

        return suites_list

    async def _read_suites(
        self,
        names: List[str],
        pattern: Optional[str] = None,
        skip_tests: Optional[str] = None,
        restore_path: Optional[str] = None,
    ) -> List[Suite]:
        """
        Read suites and return a list of Suite objects.
        """
        suites_obj = await self._get_suites_objects(names)

        await self._restore_tests(suites_obj, restore_path)

        self._filter_tests(suites_obj, pattern, False)
        self._filter_tests(suites_obj, skip_tests, True)

        num_tests = sum(len(suite_obj.tests) for suite_obj in suites_obj)
        if num_tests == 0:
            raise KirkException("No tests selected")

        if self._force_parallel:
            for suite in suites_obj:
                for test in suite.tests:
                    test.force_parallel()

        return suites_obj

    async def _exec_command(self, command: str) -> None:
        """
        Execute a single command on SUT.
        """
        async with self._exec_lock:
            exc = None
            try:
                await libkirk.events.fire("run_cmd_start", command)

                channel = self._sut.get_channel()
                test = await self._framework.find_command(channel, command)

                ret = await asyncio.wait_for(
                    channel.run_command(
                        test.full_command,
                        cwd=test.cwd,
                        env=test.env,
                        iobuffer=RedirectSUTStdout(self._sut, True),
                    ),
                    timeout=self._exec_timeout,
                )
                if not ret:
                    raise SessionError(f"Can't execute command '{test.full_command}'")

                await libkirk.events.fire(
                    "run_cmd_stop", command, ret["stdout"], ret["returncode"]
                )
            except asyncio.TimeoutError:
                exc = KirkException(f"Command timeout: {repr(command)}")
            except KirkException as err:
                if not self._stop:
                    exc = err

            if exc:
                raise exc

    async def _inner_stop(self) -> None:
        """
        Stop scheduler and SUT.
        """
        if self._scheduler:
            await self._scheduler.stop()

        await self._stop_sut()

    async def stop(self) -> None:
        """
        Stop the current session.
        """
        already_stopped = self._stop

        self._stop = True
        try:
            await self._inner_stop()

            async with self._run_lock:
                pass

            async with self._exec_lock:
                pass
        finally:
            if not already_stopped:
                await libkirk.events.fire("session_stopped")

            self._stop = False

    async def _schedule_once(self, suites_obj: List[Suite]) -> None:
        """
        Schedule tests only once.
        """
        await self._scheduler.schedule(suites_obj)
        self._results.extend(self._scheduler.results)

    async def _schedule_infinite(self, suites_obj: List[Suite]) -> None:
        """
        Schedule all testing suites infinite times.
        """
        count = 1
        while not self._stop:
            suites_iteration = []
            for suite in copy.deepcopy(suites_obj):
                suite.name = f"{suite.name}[{count}]"
                suites_iteration.append(suite)

            await self._schedule_once(suites_iteration)
            if self._scheduler.stopped:
                break

            count += 1

    async def _run_scheduler(self, suites_obj: List[Suite], runtime: float) -> None:
        """
        Run the scheduler for specific amount of time given by `runtime`.
        """
        if runtime <= 0:
            await self._schedule_once(suites_obj)
            return

        try:
            await asyncio.wait_for(self._schedule_infinite(suites_obj), runtime)
        except asyncio.TimeoutError:
            await self._scheduler.stop()

    async def _apply_fault_injection(self, fault_prob: int) -> None:
        """
        Check if we can apply fault injection configuration
        and eventually does it.
        """
        warn_msg = ""

        if not await self._sut.logged_as_root():
            if fault_prob != 0:
                warn_msg = "Run as root to use kernel fault injection"
        else:
            if await self._sut.is_fault_injection_enabled():
                await self._sut.setup_fault_injection(fault_prob)
            else:
                if fault_prob != 0:
                    warn_msg = "Fault injection is not enabled. Running tests normally"

        if warn_msg:
            self._logger.info(warn_msg)
            await libkirk.events.fire("session_warning", warn_msg)

    async def run(
        self,
        command: Optional[str] = None,
        suites: Optional[List[str]] = None,
        pattern: Optional[str] = None,
        skip_tests: Optional[str] = None,
        report_path: Optional[str] = None,
        restore_path: Optional[str] = None,
        suite_iterate: int = 1,
        randomize: bool = False,
        runtime: float = 0,
        fault_prob: int = 0,
    ) -> None:
        """
        Run a new session and store results inside a JSON file.

        :param command: Single command to run before suites.
        :type command: str
        :param suites: List of suites to execute.
        :type suites: list
        :param pattern: Regex pattern to include tests.
        :type pattern: str
        :param skip_tests: Regex for tests to skip.
        :type skip_tests: str
        :param report_path: JSON report path.
        :type report_path: str
        :param restore_path: Temporary directory generated by a previous session.
        :type restore_path: str
        :param suite_iterate: Execute all suites multiple times.
        :type suite_iterate: int
        :param randomize: Randomize all tests if True.
        :type randomize: bool
        :param runtime: For how long we want to run the session.
        :type runtime: float
        :param fault_prob: Fault injection probability.
        :type fault_prob: int
        """
        async with self._run_lock:
            await libkirk.events.fire("session_started", suites, self._tmpdir.abspath)

            channel = self._sut.get_channel()
            if not channel.parallel_execution:
                await libkirk.events.fire(
                    "session_warning", "SUT doesn't support parallel execution"
                )

            try:
                await self._start_sut()

                if command:
                    await self._exec_command(command)

                if fault_prob != 0:
                    await self._apply_fault_injection(fault_prob)

                if suites:
                    suites_obj = await self._read_suites(
                        suites, pattern, skip_tests, restore_path
                    )

                    suites_obj = self._apply_iterate(suites_obj, suite_iterate)

                    if randomize:
                        for suite in suites_obj:
                            random.shuffle(suite.tests)

                    await self._run_scheduler(suites_obj, runtime)
            except KirkException as err:
                if not self._stop:
                    self._logger.exception(err)
                    await libkirk.events.fire("session_error", str(err))
                    raise err
            finally:
                try:
                    # configure fault injection to the original values
                    if fault_prob != 0:
                        await self._apply_fault_injection(0)

                    if self._results:
                        exporter = JSONExporter()

                        tasks = [
                            exporter.save_file(
                                self._results,
                                os.path.join(self._tmpdir.abspath, "results.json"),
                            )
                        ]

                        if report_path:
                            tasks.append(exporter.save_file(self._results, report_path))

                        await asyncio.gather(*tasks)
                except KirkException as err:
                    self._logger.exception(err)
                    await libkirk.events.fire("session_error", str(err))
                    raise err
                finally:
                    self._results.clear()
                    await self._inner_stop()

                    await libkirk.events.fire(
                        "session_completed", self._scheduler.results
                    )



"""
.. module:: ltp
    :platform: Linux
    :synopsis: LTP framework definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import json
import logging
import os
import re
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from libkirk.com import ComChannel
from libkirk.data import (
    Suite,
    Test,
)
from libkirk.errors import FrameworkError
from libkirk.framework import Framework
from libkirk.results import (
    ResultStatus,
    TestResults,
)

# Mapping from LTP return code to ResultStatus
_RETCODE_STATUS: Dict[int, int] = {
    0: ResultStatus.PASS,
    2: ResultStatus.BROK,
    -1: ResultStatus.BROK,
    4: ResultStatus.WARN,
    32: ResultStatus.CONF,
}

_ANSI_ESCAPE = re.compile(r"\u001b\[[0-9;]+[a-zA-Z]")

_SUMMARY_RE = re.compile(
    r"Summary:\n"
    r"passed\s*(?P<passed>\d+)\n"
    r"failed\s*(?P<failed>\d+)\n"
    r"broken\s*(?P<broken>\d+)\n"
    r"skipped\s*(?P<skipped>\d+)\n"
    r"warnings\s*(?P<warnings>\d+)\n",
)


class LTPFramework(Framework):
    """
    Linux Test Project framework definition.
    """

    # Tags whose presence marks a test as non-parallelizable.
    PARALLEL_BLACKLIST: frozenset = frozenset(
        {
            "needs_root",
            "needs_device",
            "mount_device",
            "mntpoint",
            "resource_file",
            "format_device",
            "save_restore",
            "max_runtime",
        }
    )

    # Environment variables without the `LTP_` or `TST_` prefix that are still forwarded.
    SUPPORTED_ENV: frozenset = frozenset(
        {
            "PATH",
            "KCONFIG_PATH",
            "KCONFIG_SKIP_CHECK",
            # LTP network test variables
            "RHOST",
            "IPV4_LHOST",
            "IPV4_RHOST",
            "IPV6_LHOST",
            "IPV6_RHOST",
            "LHOST_IFACES",
            "RHOST_IFACES",
            # Stress test parameters
            "NS_DURATION",
            "NS_TIMES",
            "CONNECTION_TOTAL",
            "IP_TOTAL",
            "IP_TOTAL_FOR_TCPIP",
            "ROUTE_TOTAL",
            "ROUTE_CHANGE_IP",
            "ROUTE_CHANGE_NETLINK",
            "MTU_CHANGE_TIMES",
            "IF_UPDOWN_TIMES",
            "PING_MAX",
            # ICMP / multicast
            "NS_ICMPV4_SENDER_DATA_MAXSIZE",
            "NS_ICMPV6_SENDER_DATA_MAXSIZE",
            "MCASTNUM_NORMAL",
            "MCASTNUM_HEAVY",
            # File transfer / services
            "DOWNLOAD_BIGFILESIZE",
            "DOWNLOAD_REGFILESIZE",
            "UPLOAD_BIGFILESIZE",
            "UPLOAD_REGFILESIZE",
            "HTTP_DOWNLOAD_DIR",
            "FTP_DOWNLOAD_DIR",
            "FTP_UPLOAD_DIR",
            "FTP_UPLOAD_URLDIR",
            # IPsec / virtual interfaces
            "IPSEC_MODE",
            "IPSEC_PROTO",
            "VIRT_PERF_THRESHOLD",
        }
    )

    # Variables set explicitly in _update_env_vars; skip them in the loop.
    _PRESET_ENV: frozenset = frozenset(
        {
            "LTPROOT",
            "TMPDIR",
            "LTP_COLORIZE_OUTPUT",
            "LTP_TIMEOUT_MUL",
        }
    )

    def __init__(
        self,
        max_runtime: float = 0.0,
        timeout: float = 30.0,
    ) -> None:
        """
        :param max_runtime: filter out all tests above this time value
        :type max_runtime: float
        :param timeout: generic tests timeout
        :type timeout: float
        """
        self._logger = logging.getLogger("libkirk.ltp")
        self._cmd_matcher = re.compile(r'(?:"[^"]*"|\'[^\']*\'|\S+)')
        self._max_runtime = max_runtime
        self._root = os.environ.get("LTPROOT", "/opt/ltp")
        self._tc_folder = os.path.join(self._root, "testcases", "bin")
        self._env: Dict[str, str] = {}

        self._update_env_vars(timeout)

    def _update_env_vars(self, timeout: float) -> None:
        """
        Populate self._env with LTP-relevant environment variables.
        """
        self._env["LTPROOT"] = self._root
        self._env["TMPDIR"] = os.environ.get("TMPDIR", "/tmp")
        self._env["LTP_COLORIZE_OUTPUT"] = os.environ.get("LTP_COLORIZE_OUTPUT", "1")

        multiplier = os.environ.get("LTP_TIMEOUT_MUL")
        if multiplier:
            self._env["LTP_TIMEOUT_MUL"] = multiplier
        elif timeout:
            self._env["LTP_TIMEOUT_MUL"] = str((timeout * 0.9) / 300.0)

        for key, val in os.environ.items():
            if key in self._PRESET_ENV:
                continue

            if key in self.SUPPORTED_ENV or key.startswith("LTP_") or key.startswith("TST_"):
                self._env[key] = val

    async def _read_path(self, channel: ComChannel) -> Dict[str, str]:
        """
        Return a copy of self._env with the testcases folder appended to PATH.
        """
        env = self._env.copy()
        if "PATH" in env:
            env["PATH"] = f"{env['PATH']}:{self._tc_folder}"
        else:
            ret = await channel.run_command("echo -n $PATH")

            if not ret or ret["returncode"] != 0:
                raise FrameworkError("Can't read PATH variable")

            env["PATH"] = f"{ret['stdout'].strip()}:{self._tc_folder}"

        self._logger.debug("PATH=%s", env["PATH"])

        return env

    def _is_addable(self, test_params: Dict[str, Any]) -> bool:
        """
        Return False when max_runtime filtering is active and the test exceeds
        the configured limit.
        """
        if not self._max_runtime:
            return True

        runtime = test_params.get("max_runtime")
        if runtime is None:
            return True

        try:
            if float(runtime) >= self._max_runtime:
                self._logger.info("max_runtime is bigger than %f", self._max_runtime)

                return False
        except TypeError:
            self._logger.error("metadata contains wrong max_runtime type: %s", runtime)

        return True

    def _get_cmd_args(self, line: str) -> List[str]:
        """
        Split a runtest line into a list of command + arguments.

        Handles quoted arguments, e.g.::

            cmd -c "cmd2 -g arg1 -t arg2"
        """
        return self._cmd_matcher.findall(line)

    async def _read_runtest(
        self,
        channel: ComChannel,
        suite_name: str,
        content: str,
        metadata: Optional[dict] = None,
    ) -> Suite:
        """
        Parse a runtest file and return the corresponding Suite.
        """
        self._logger.info("collecting testing suite: %s", suite_name)

        metadata_tests: Optional[dict] = None
        if metadata:
            self._logger.info("Reading metadata content")
            metadata_tests = metadata.get("tests")

        env = await self._read_path(channel)
        tests: List[Test] = []

        for line in content.split("\n"):
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            self._logger.debug("Test declaration: %s", line)

            parts = self._get_cmd_args(line)
            if len(parts) < 2:
                raise FrameworkError("runtest file is not defining test command")

            test_name, test_cmd, *test_args = parts
            parallelizable = False

            if metadata_tests is not None:
                test_params = metadata_tests.get(test_name)
                if test_params is None:
                    # Test not using the new LTP API – parallelism unknown.
                    self._logger.info("Found %s test params in metadata", test_name)
                else:
                    self._logger.info("Found %s test params in metadata", test_name)
                    self._logger.debug("params=%s", test_params)

                    if not self._is_addable(test_params):
                        continue

                    parallelizable = not (self.PARALLEL_BLACKLIST & test_params.keys())

            self._logger.info(
                "Test '%s' is%s parallelizable",
                test_name,
                "" if parallelizable else " not",
            )

            test = Test(
                name=test_name,
                cmd=test_cmd,
                args=test_args,
                cwd=self._tc_folder,
                env=env,
                parallelizable=parallelizable,
            )
            tests.append(test)

            self._logger.debug("test: %s", test)

        self._logger.debug("Collected tests: %d", len(tests))

        suite = Suite(suite_name, tests)
        self._logger.debug(suite)
        self._logger.info("Collected testing suite: %s", suite_name)

        return suite

    async def get_suites(self, channel: ComChannel) -> List[str]:
        if not channel:
            raise ValueError("SUT is None")

        ret = await channel.run_command(f"test -d {self._root}")
        if not ret or ret["returncode"] != 0:
            raise FrameworkError(f"LTP folder doesn't exist: {self._root}")

        runtest_dir = os.path.join(self._root, "runtest")
        ret = await channel.run_command(f"test -d {runtest_dir}")
        if not ret or ret["returncode"] != 0:
            raise FrameworkError(f"'{runtest_dir}' doesn't exist inside SUT")

        ret = await channel.run_command(f"ls --format=single-column {runtest_dir}")
        if not ret:
            raise FrameworkError("Can't communicate with SUT")

        stdout = ret["stdout"]
        if ret["returncode"] != 0:
            raise FrameworkError(f"command failed with: {stdout}")

        return [line for line in stdout.split("\n") if line]

    async def find_command(self, channel: ComChannel, command: str) -> Test:
        if not channel:
            raise ValueError("SUT is None")
        if not command:
            raise ValueError("command is empty")

        cmd_args = self._get_cmd_args(command)
        cwd = None
        env = None

        ret = await channel.run_command(f"test -d {self._tc_folder}")
        if ret and ret["returncode"] == 0:
            cwd = self._tc_folder
            env = await self._read_path(channel)

        return Test(
            name=cmd_args[0],
            cmd=cmd_args[0],
            args=cmd_args[1:] or None,
            cwd=cwd,
            env=env,
            parallelizable=False,
        )

    async def find_suite(self, channel: ComChannel, name: str) -> Suite:
        if not channel:
            raise ValueError("SUT is None")
        if not name:
            raise ValueError("name is empty")

        ret = await channel.run_command(f"test -d {self._root}")
        if not ret or ret["returncode"] != 0:
            raise FrameworkError(f"LTP folder doesn't exist: {self._root}")

        suite_path = os.path.join(self._root, "runtest", name)
        ret = await channel.run_command(f"test -f {suite_path}")
        if not ret or ret["returncode"] != 0:
            raise FrameworkError(f"'{name}' suite doesn't exist")

        runtest_str = (await channel.fetch_file(suite_path)).decode(
            encoding="utf-8", errors="ignore"
        )

        metadata_dict = None
        metadata_path = os.path.join(self._root, "metadata", "ltp.json")
        ret = await channel.run_command(f"test -f {metadata_path}")
        if ret and ret["returncode"] == 0:
            metadata_dict = json.loads(await channel.fetch_file(metadata_path))

        return await self._read_runtest(channel, name, runtest_str, metadata_dict)

    async def read_result(
        self, test: Test, stdout: str, retcode: int, exec_t: float
    ) -> TestResults:
        stdout = _ANSI_ESCAPE.sub("", stdout)

        match = _SUMMARY_RE.search(stdout)
        if match:
            passed = int(match.group("passed"))
            failed = int(match.group("failed"))
            broken = int(match.group("broken"))
            skipped = int(match.group("skipped"))
            warnings = int(match.group("warnings"))
        else:
            passed = stdout.count("TPASS")
            failed = stdout.count("TFAIL")
            skipped = stdout.count("TSKIP")
            broken = stdout.count("TBROK")
            warnings = stdout.count("TWARN")

            if not any((passed, failed, skipped, broken, warnings)):
                # Legacy test: derive counts from the return code alone.
                if retcode == 0:
                    passed = 1
                elif retcode == 4:
                    warnings = 1
                elif retcode == 32:
                    skipped = 1
                elif retcode != -1:
                    failed = 1

        status = _RETCODE_STATUS.get(retcode, ResultStatus.FAIL)

        if retcode == -1:
            broken = 1

        return TestResults(
            test=test,
            failed=failed,
            passed=passed,
            broken=broken,
            skipped=skipped,
            warnings=warnings,
            exec_time=exec_t,
            retcode=retcode,
            stdout=stdout,
            status=status,
        )



"""
.. module:: qemu
    :platform: Linux
    :synopsis: module containing qemu channel implementation

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import contextlib
import logging
import os
import re
import secrets
import shutil
import signal
import string
import time
from typing import (
    Any,
    Dict,
    Optional,
)

import libkirk.types
from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import (
    CommunicationError,
    KernelPanicError,
)
from libkirk.io import AsyncFile


class QemuComChannel(ComChannel):
    """
    Qemu com channel spawn a new VM using qemu and execute commands inside it.
    """

    _name = "qemu"

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.qemu")
        self._comm_lock = asyncio.Lock()
        self._cmd_lock = asyncio.Lock()
        self._fetch_lock = asyncio.Lock()
        self._tmpdir = ""
        self._proc = None
        self._stop = False
        self._logged_in = False
        self._last_pos = 0
        self._user = None
        self._password = None
        self._prompt = "#"
        self._ram = None
        self._smp = None
        self._virtfs = None
        self._serial_type = None
        self._qemu_cmd = ""
        self._opts = None
        self._image = None
        self._kernel = None
        self._initrd = None
        self._last_read = ""
        self._panic = False

    @staticmethod
    def _generate_string(length: int = 10) -> str:
        """
        Generate a random string of the given length.
        """
        out = "".join(
            secrets.choice(string.ascii_letters + string.digits) for _ in range(length)
        )
        return out

    def _get_transport(self) -> tuple:
        """
        Return a couple of transport_dev and transport_file used by
        qemu instance for transport configuration.
        """
        pid = os.getpid()
        transport_file = os.path.join(self._tmpdir, f"transport-{pid}")
        transport_dev = ""

        if self._serial_type == "isa":
            transport_dev = "/dev/ttyS1"
        elif self._serial_type == "virtio":
            transport_dev = "/dev/vport1p1"

        return transport_dev, transport_file

    def _get_command(self) -> str:
        """
        Return the full qemu command to execute.
        """
        pid = os.getpid()
        tty_log = os.path.join(self._tmpdir, f"ttyS0-{pid}.log")

        params = []
        params.append("-enable-kvm")
        params.append("-display none")
        params.append(f"-m {self._ram}")
        params.append(f"-smp {self._smp}")
        params.append("-device virtio-rng-pci")
        params.append(f"-chardev stdio,id=tty,logfile={tty_log}")

        if self._serial_type == "isa":
            params.append("-serial chardev:tty")
            params.append("-serial chardev:transport")
        elif self._serial_type == "virtio":
            params.append("-device virtio-serial")
            params.append("-device virtconsole,chardev=tty")
            params.append("-device virtserialport,chardev=transport")
        else:
            raise NotImplementedError(
                f"Unsupported serial device type {self._serial_type}"
            )

        _, transport_file = self._get_transport()
        params.append(f"-chardev file,id=transport,path={transport_file}")

        if self._virtfs:
            params.append(
                "-virtfs local,"
                f"path={self._virtfs},"
                "mount_tag=host0,"
                "security_model=mapped-xattr,"
                "readonly=on"
            )

        if self._image:
            params.append(f"-drive if=virtio,cache=unsafe,file={self._image}")

        if self._initrd:
            params.append(f"-initrd {self._initrd}")

        if self._kernel:
            console = "ttyS0"
            if self._serial_type == "virtio":
                console = "hvc0"

            params.append(f"-append 'console={console} ignore_loglevel'")
            params.append(f"-kernel {self._kernel}")

        if self._opts:
            params.append(self._opts)

        cmd = f"{self._qemu_cmd} {' '.join(params)}"

        return cmd

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        self._logger.info("Initialize Qemu")

        self._tmpdir = libkirk.types.dict_item(kwargs, "tmpdir", str, None)
        self._user = libkirk.types.dict_item(kwargs, "user", str, None)
        self._password = libkirk.types.dict_item(kwargs, "password", str, None)
        self._prompt = libkirk.types.dict_item(kwargs, "prompt", str, "#")
        self._image = libkirk.types.dict_item(kwargs, "image", str, None)
        self._initrd = libkirk.types.dict_item(kwargs, "initrd", str, None)
        self._kernel = libkirk.types.dict_item(kwargs, "kernel", str, None)
        self._ram = libkirk.types.dict_item(kwargs, "ram", str, "2G")
        self._smp = libkirk.types.dict_item(kwargs, "smp", str, "2")
        self._virtfs = libkirk.types.dict_item(kwargs, "virtfs", str, None)
        self._serial_type = libkirk.types.dict_item(kwargs, "serial", str, "isa")
        self._opts = libkirk.types.dict_item(kwargs, "options", str, None)

        system = libkirk.types.dict_item(kwargs, "system", str, "x86_64")
        self._qemu_cmd = f"qemu-system-{system}"

        if not self._tmpdir or not os.path.isdir(self._tmpdir):
            raise CommunicationError(
                f"Temporary directory doesn't exist: {self._tmpdir}"
            )

        if self._image and not os.path.isfile(self._image):
            raise CommunicationError(f"Image location doesn't exist: {self._image}")

        if self._kernel and not os.path.isfile(self._kernel):
            raise CommunicationError(f"Kernel location doesn't exist: {self._kernel}")

        if self._initrd and not os.path.isfile(self._initrd):
            raise CommunicationError(f"initrd location doesn't exist: {self._initrd}")

        if not self._ram:
            raise CommunicationError("RAM is not defined")

        if not self._smp:
            raise CommunicationError("CPU is not defined")

        if self._virtfs and not os.path.isdir(self._virtfs):
            raise CommunicationError(
                f"Virtual FS directory doesn't exist: {self._virtfs}"
            )

        if self._serial_type not in ["isa", "virtio"]:
            raise CommunicationError("Serial protocol must be isa or virtio")

    @property
    def config_help(self) -> Dict[str, str]:
        return {
            "image": "qemu image location",
            "kernel": "kernel image location",
            "initrd": "initrd image location",
            "user": "user name (default: '')",
            "password": "user password (default: '')",
            "prompt": "prompt string (default: '#')",
            "system": "system architecture (default: x86_64)",
            "ram": "RAM of the VM (default: 2G)",
            "smp": "number of CPUs (default: 2)",
            "serial": "type of serial protocol. isa|virtio (default: isa)",
            "virtfs": "directory to mount inside VM",
            "options": "user defined options",
        }

    @property
    def parallel_execution(self) -> bool:
        return False

    async def active(self) -> bool:
        if self._proc is None:
            return False

        with contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(self._proc.wait(), 1e-6)

        return self._proc.returncode is None

    async def ping(self) -> float:
        if not await self.active():
            raise CommunicationError("Qemu is not running")

        _, _, exec_time = await self._exec("test .", None)

        return exec_time

    async def _read_stdout(self, size: int, iobuffer: Optional[IOBuffer] = None) -> str:
        """
        Read data from stdout.
        """
        # pyrefly: ignore[missing-attribute]
        data = await self._proc.stdout.read(size)
        rdata = data.decode(encoding="utf-8", errors="replace")

        # write on stdout buffers
        if iobuffer:
            await iobuffer.write(rdata)

        return rdata

    async def _write_stdin(self, data: str) -> None:
        """
        Write data on stdin.
        """
        if not await self.active():
            return

        wdata = data.encode(encoding="utf-8")
        try:
            # pyrefly: ignore[missing-attribute]
            self._proc.stdin.write(wdata)
        except BrokenPipeError as err:
            if not self._stop:
                raise CommunicationError(err) from err

    async def _wait_for(
        self, message: str, iobuffer: Optional[IOBuffer] = None
    ) -> Optional[str]:
        """
        Wait a string from stdout.
        """
        if not await self.active():
            return None

        self._logger.info("Waiting for message: %s", repr(message))

        stdout = self._last_read
        self._panic = False

        while True:
            if self._stop or self._panic:
                break

            if not await self.active():
                break

            message_pos = stdout.find(message)
            if message_pos != -1:
                self._last_read = stdout[message_pos + len(message) :]
                break

            data = await self._read_stdout(1024, iobuffer)
            if data:
                stdout += data

            if "Kernel panic" in stdout:
                # give time to panic message coming out from serial
                await asyncio.sleep(2)

                # read as much data as possible from stdout
                data = await self._read_stdout(1024 * 1024, iobuffer)
                stdout += data

                self._panic = True

        if self._panic:
            # if we ended before raising Kernel panic, we raise the exception
            raise KernelPanicError()

        return stdout

    async def _wait_lockers(self) -> None:
        """
        Wait for Qemu lockers to be released.
        """
        async with self._comm_lock:
            pass

        async with self._cmd_lock:
            pass

        async with self._fetch_lock:
            pass

    async def _exec(self, command: str, iobuffer: Optional[IOBuffer] = None) -> tuple:
        """
        Execute a command and return set(stdout, retcode, exec_time).
        """
        self._logger.debug("Execute command: %s", repr(command))

        code = self._generate_string()

        msg = f"echo $?-{code}\n"
        if command and command.rstrip():
            msg = f"{command};" + msg

        self._logger.info("Sending %s", repr(msg))

        t_start = time.time()

        await self._write_stdin(msg)
        stdout = await self._wait_for(code, iobuffer)

        exec_time = time.time() - t_start

        retcode = -1

        if not self._stop:
            if stdout and stdout.rstrip():
                match = re.search(f"(?P<retcode>\\d+)-{code}", stdout)
                if not match:
                    raise CommunicationError(
                        f"Can't read return code from reply {repr(stdout)}"
                    )

                # first character is '\n'
                stdout = stdout[1 : match.start()]

                try:
                    retcode = int(match.group("retcode"))
                except TypeError:
                    pass

        self._logger.debug(
            "stdout=%s, retcode=%d, exec_time=%d", repr(stdout), retcode, exec_time
        )

        return stdout, retcode, exec_time

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not await self.active():
            return

        self._logger.info("Shutting down virtual machine")
        self._stop = True

        try:
            if not self._panic:
                # stop command first
                if self._cmd_lock.locked() or self._fetch_lock.locked():
                    self._logger.info("Stop running command")

                    # send interrupt character (equivalent of CTRL+C)
                    await self._write_stdin("\x03")
                    await self._wait_lockers()

                # logged in -> poweroff
                if self._logged_in:
                    self._logger.info("Poweroff virtual machine")

                    await self._write_stdin("poweroff; poweroff -f\n")

                    while await self.active():
                        await self._read_stdout(1024, iobuffer)

                    # pyrefly: ignore[missing-attribute]
                    await self._proc.wait()
        except asyncio.TimeoutError:
            pass
        finally:
            # still running -> stop process
            if await self.active():
                self._logger.info("Killing virtual machine")

                # pyrefly: ignore[missing-attribute]
                self._proc.kill()

                await self._wait_lockers()
                # pyrefly: ignore[missing-attribute]
                await self._proc.wait()

            self._stop = False

        self._logger.info("Qemu process ended")

    async def communicate(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not shutil.which(self._qemu_cmd):
            raise CommunicationError(f"Command not found: {self._qemu_cmd}")

        if await self.active():
            raise CommunicationError("Virtual machine is already running")

        error = None

        async with self._comm_lock:
            self._logged_in = False

            cmd = self._get_command()

            self._logger.info("Starting virtual machine")
            self._logger.debug(cmd)

            # pyrefly: ignore[bad-assignment]
            self._proc = await asyncio.create_subprocess_shell(
                cmd,
                stdout=asyncio.subprocess.PIPE,
                stdin=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )

            try:
                if self._user:
                    await self._wait_for("login:", iobuffer)
                    await self._write_stdin(f"{self._user}\n")

                    if self._password:
                        await self._wait_for("Password:", iobuffer)
                        await self._write_stdin(f"{self._password}\n")

                    await asyncio.sleep(0.2)

                await self._wait_for(self._prompt, iobuffer)
                await asyncio.sleep(0.2)

                await self._write_stdin("stty -echo; stty cols 1024\n")
                await self._wait_for(self._prompt, None)

                await self._write_stdin("dmesg -D\n")
                await self._wait_for(self._prompt, None)

                _, retcode, _ = await self._exec("export PS1=''", None)
                if retcode != 0:
                    raise CommunicationError("Can't setup prompt string")

                if self._virtfs:
                    _, retcode, _ = await self._exec(
                        "mount -t 9p -o trans=virtio host0 /mnt", None
                    )
                    if retcode != 0:
                        raise CommunicationError("Failed to mount virtfs")

                self._logged_in = True

                self._logger.info("Virtual machine started")
            except CommunicationError as err:
                error = err

        if not self._stop and error:
            # this can happen when shell is available but
            # something happened during commands execution
            await self.stop(iobuffer=iobuffer)

            raise CommunicationError(error)

    async def run_command(
        self,
        command: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        iobuffer: Optional[IOBuffer] = None,
    ) -> Optional[Dict[str, Any]]:
        if not command:
            raise ValueError("command is empty")

        if not await self.active():
            raise CommunicationError("Virtual machine is not running")

        async with self._cmd_lock:
            self._logger.info("Running command: %s", command)

            if cwd:
                stdout, retcode, _ = await self._exec(f"cd {cwd}", None)
                if retcode != 0:
                    raise CommunicationError(
                        f"Can't setup current working directory: {stdout}"
                    )

            if env:
                for key, value in env.items():
                    stdout, retcode, _ = await self._exec(f"export {key}={value}", None)
                    if retcode != 0:
                        raise CommunicationError(
                            f"Can't setup env {key}={value}: {stdout}"
                        )

            stdout, retcode, exec_time = await self._exec(f"{command}", iobuffer)

            ret = {
                "command": command,
                "returncode": retcode,
                "stdout": stdout,
                "exec_time": exec_time,
            }

            self._logger.debug(ret)

            return ret

    async def fetch_file(self, target_path: str) -> bytes:
        if not target_path:
            raise ValueError("target path is empty")

        if not await self.active():
            raise CommunicationError("Virtual machine is not running")

        async with self._fetch_lock:
            self._logger.info("Downloading %s", target_path)

            _, retcode, _ = await self._exec(f"test -f {target_path}", None)
            if retcode != 0:
                raise CommunicationError(f"'{target_path}' doesn't exist")

            transport_dev, transport_path = self._get_transport()

            stdout, retcode, _ = await self._exec(
                f"cat {target_path} > {transport_dev}", None
            )

            if self._stop:
                return bytes()

            if retcode not in [0, signal.SIGHUP, signal.SIGKILL]:
                raise CommunicationError(
                    f"Can't send file to {transport_dev}: {stdout}"
                )

            # read back data and send it to the local file path
            file_size = os.path.getsize(transport_path)

            retdata = bytearray()

            async with AsyncFile(transport_path, "rb") as transport:
                while not self._stop and self._last_pos < file_size:
                    await transport.seek(self._last_pos)

                    data = await transport.read(4096)
                    if data:
                        # pyrefly: ignore[bad-argument-type]
                        retdata.extend(data)

                    pos = await transport.tell()
                    if pos is None:
                        raise CommunicationError("Can't read file position")

                    self._last_pos = pos

            self._logger.info("File downloaded")

            return bytes(retdata)



"""
.. module:: results
    :platform: Linux
    :synopsis: module containing suites data definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import (
    List,
    Optional,
)

from libkirk.data import (
    Suite,
    Test,
)


class ResultStatus:
    """
    Overall status of the test. This is a specific flag that is used to
    recognize final test status. For example, we might have 10 tests passing
    inside a single test binary, but the overall status of the test is fine, so
    we assign a PASS status.
    """

    PASS = 0
    """
    Test has passed.
    """

    BROK = 2
    """
    Test is broken.
    """

    WARN = 4
    """
    Test warnings received.
    """

    FAIL = 16
    """
    Test has failed.
    """

    CONF = 32
    """
    Test can't run because of configuration error.
    """


class Results:
    """
    Base class for results.
    """

    @property
    def exec_time(self) -> float:
        """
        :return: Test execution time in seconds.
        :rtype: float
        """
        raise NotImplementedError()

    @property
    def failed(self) -> int:
        """
        :return: Number of failures.
        :rtype: int
        """
        raise NotImplementedError()

    @property
    def passed(self) -> int:
        """
        :return: Number of passed tests.
        :rtype: int
        """
        raise NotImplementedError()

    @property
    def broken(self) -> int:
        """
        :return: Number of broken tests.
        :rtype: int
        """
        raise NotImplementedError()

    @property
    def skipped(self) -> int:
        """
        :return: Number of skipped tests.
        :rtype: int
        """
        raise NotImplementedError()

    @property
    def warnings(self) -> int:
        """
        :return: Number of warnings.
        :rtype: int
        """
        raise NotImplementedError()


class TestResults(Results):
    """
    Test results definition.
    """

    def __init__(
        self,
        test: Test,
        failed: int = 0,
        passed: int = 0,
        broken: int = 0,
        skipped: int = 0,
        warnings: int = 0,
        exec_time: float = 0.0,
        status: int = ResultStatus.PASS,
        retcode: int = 0,
        stdout: str = "",
    ) -> None:
        """
        :param test: Test object declaration.
        :type test: Test
        :param failed: number of TFAIL.
        :type failed: int
        :param passed: number of TPASS.
        :type passed: int
        :param broken: number of TBROK.
        :type broken: int
        :param skipped: number of TSKIP.
        :type skipped: int
        :param warnings: number of TWARN.
        :type warnings: int
        :param exec_time: Time for test's execution.
        :type exec_time: float
        :param status: Overall status of the test.
        :type status: int
        :param retcode: Return code of the executed test.
        :type retcode: int
        :param stdout: Stdout of the test.
        :type stdout: str
        """
        if not test:
            raise ValueError("Empty test object")

        self._test = test
        self._failed = failed
        self._passed = passed
        self._broken = broken
        self._skipped = skipped
        self._warns = warnings
        self._exec_time = exec_time
        self._status = status
        self._retcode = retcode
        self._stdout = stdout

    def __repr__(self) -> str:
        return (
            f"test: '{self._test}', "
            f"failed: '{self._failed}', "
            f"passed: {self._passed}, "
            f"broken: {self._broken}, "
            f"skipped: {self._skipped}, "
            f"warnings: {self._warns}, "
            f"exec_time: {self._exec_time}, "
            f"status: {self._status}, "
            f"retcode: {self._retcode}, "
            f"stdout: {repr(self._stdout)}"
        )

    @property
    def test(self) -> Test:
        """
        :return: Test object.
        :rtype: Test
        """
        return self._test

    @property
    def return_code(self) -> int:
        """
        :return: Return code after execution.
        :rtype: int
        """
        return self._retcode

    @property
    def stdout(self) -> str:
        """
        :return: Test process stdout.
        :rtype: str
        """
        return self._stdout

    @property
    def status(self) -> int:
        """
        :return: Test result status.
        :rtype: int
        """
        return self._status

    @property
    def exec_time(self) -> float:
        return self._exec_time

    @property
    def failed(self) -> int:
        return self._failed

    @property
    def passed(self) -> int:
        return self._passed

    @property
    def broken(self) -> int:
        return self._broken

    @property
    def skipped(self) -> int:
        return self._skipped

    @property
    def warnings(self) -> int:
        return self._warns


class SuiteResults(Results):
    """
    Testing suite results definition.
    """

    def __init__(
        self,
        suite: Suite,
        tests: Optional[List[TestResults]] = None,
        distro: Optional[str] = None,
        distro_ver: Optional[str] = None,
        kernel: Optional[str] = None,
        cmdline: Optional[str] = None,
        arch: Optional[str] = None,
        cpu: Optional[str] = None,
        swap: Optional[str] = None,
        ram: Optional[str] = None,
    ) -> None:
        """
        :param suite: Test object declaration.
        :type suite: Suite
        :param tests: List of the tests results.
        :type tests: list(TestResults)
        :param distro: Distribution name.
        :type distro: str
        :param distro_ver: Distribution version.
        :type distro_ver: str
        :param kernel: Kernel version.
        :type kernel: str
        :param cmdline: /proc/cmdline.
        :type cmdline: str
        :param arch: OS architecture.
        :type arch: str
        :param cpu: CPU type info.
        :type cpu: str
        :param swap: Swap memory info.
        :type swap: str
        :param ram: RAM info.
        :type ram: str
        """
        if not suite:
            raise ValueError("Empty suite object")

        self._suite = suite
        self._tests = tests if tests is not None else []
        self._distro = distro
        self._distro_ver = distro_ver
        self._kernel = kernel
        self._cmdline = cmdline
        self._arch = arch
        self._cpu = cpu
        self._swap = swap
        self._ram = ram

    def __repr__(self) -> str:
        return (
            f"suite: '{self._suite}', "
            f"tests: '{self._tests}', "
            f"distro: {self._distro}, "
            f"distro_ver: {self._distro_ver}, "
            f"kernel: {self._kernel}, "
            f"cmdline: {self._cmdline}, "
            f"arch: {self._arch}, "
            f"cpu: {self._cpu}, "
            f"swap: {self._swap}, "
            f"ram: {self._ram}"
        )

    @property
    def suite(self) -> Suite:
        """
        :returns: Testing suite.
        :rtype: Suite
        """
        return self._suite

    @property
    def tests_results(self) -> List[TestResults]:
        """
        :returns: list of all tests results.
        :rtype: list(TestResults)
        """
        return self._tests

    def _get_result(self, attr: str) -> int:
        """
        Return the total number of results.
        """
        res = 0
        for test in self._tests:
            res += getattr(test, attr)

        return res

    @property
    def distro(self) -> Optional[str]:
        """
        :return: Linux distribution name.
        :rtype: str | None
        """
        return self._distro

    @property
    def distro_ver(self) -> Optional[str]:
        """
        :return: Linux distribution version.
        :rtype: str | None
        """
        return self._distro_ver

    @property
    def kernel(self) -> Optional[str]:
        """
        :return: Kernel version.
        :rtype: str | None
        """
        return self._kernel

    @property
    def cmdline(self) -> Optional[str]:
        """
        :return: /proc/cmdline.
        :rtype: str | None
        """
        return self._cmdline

    @property
    def arch(self) -> Optional[str]:
        """
        :return: Operating system architecture.
        :rtype: str | None
        """
        return self._arch

    @property
    def cpu(self) -> Optional[str]:
        """
        :return: Current CPU type.
        :rtype: str | None
        """
        return self._cpu

    @property
    def swap(self) -> Optional[str]:
        """
        :return: Current swap memory occupation.
        :rtype: str | None
        """
        return self._swap

    @property
    def ram(self) -> Optional[str]:
        """
        :return: Current RAM occupation.
        :rtype: str | None
        """
        return self._ram

    @property
    def exec_time(self) -> float:
        return self._get_result("exec_time")

    @property
    def failed(self) -> int:
        return self._get_result("failed")

    @property
    def passed(self) -> int:
        return self._get_result("passed")

    @property
    def broken(self) -> int:
        return self._get_result("broken")

    @property
    def skipped(self) -> int:
        return self._get_result("skipped")

    @property
    def warnings(self) -> int:
        return self._get_result("warnings")



"""
.. module:: ltx_chan
    :platform: Linux
    :synopsis: module containing LTX communication channel

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import importlib.util
import logging
import os
import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import libkirk.types
from libkirk.channels.ltx import (
    LTX,
    Request,
    Requests,
)
from libkirk.com import (
    ComChannel,
    IOBuffer,
)
from libkirk.errors import (
    CommunicationError,
    LTXError,
)


class LTXComChannel(ComChannel):
    """
    Communication channel using LTX as executor.
    """

    _name = "ltx"

    def __init__(self) -> None:
        self._logger = logging.getLogger("kirk.ltx")
        self._release_lock = asyncio.Lock()
        self._fetch_lock = asyncio.Lock()
        self._ltx: Optional[LTX] = None
        self._outfile = ""
        self._infile = ""
        self._slots = []

    @property
    def config_help(self) -> Dict[str, str]:
        return {
            "infile": "file where ltx is reading data",
            "outfile": "file where ltx is writing data",
        }

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        if not importlib.util.find_spec("msgpack"):
            raise CommunicationError("'msgpack' library is not available")

        self._logger.info("Initialize LTX channel")

        self._infile = libkirk.types.dict_item(kwargs, "infile", str)
        self._outfile = libkirk.types.dict_item(kwargs, "outfile", str)

        if not self._infile or not os.path.exists(self._infile):
            raise CommunicationError(f"'{self._infile}' input file doesn't exist")

        if not self._outfile or not os.path.exists(self._outfile):
            raise CommunicationError(f"'{self._outfile}' output file doesn't exist")

    @property
    def parallel_execution(self) -> bool:
        return True

    async def active(self) -> bool:
        if not self._ltx:
            return False

        return self._ltx.connected

    async def stop(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if not await self.active():
            return

        if self._slots:
            requests = []
            for slot_id in self._slots:
                requests.append(Requests.kill(slot_id))

            if requests:
                await self._send_requests(requests)

                while self._slots:
                    await asyncio.sleep(1e-2)

        try:
            # pyrefly: ignore[missing-attribute]
            await self._ltx.disconnect()
        except LTXError as err:
            raise CommunicationError(err) from err

        while await self.active():
            await asyncio.sleep(1e-2)

    async def _send_requests(self, requests: List[Request]) -> Dict[Request, Any]:
        """
        Send requests and check for LTXError.
        """
        reply = None
        try:
            # pyrefly: ignore[missing-attribute]
            reply = await self._ltx.gather(requests)
        except LTXError as err:
            raise CommunicationError(err) from err

        return reply

    async def _reserve_slot(self) -> int:
        """
        Reserve an execution slot.
        """
        async with self._release_lock:
            slot_id = -1
            for i in range(0, Request.MAX_SLOTS):
                if i not in self._slots:
                    slot_id = i
                    break

            if slot_id == -1:
                raise CommunicationError("No execution slots available")

            self._slots.append(slot_id)

            return slot_id

    async def _release_slot(self, slot_id: int) -> None:
        """
        Release an execution slot.
        """
        async with self._release_lock:
            if slot_id in self._slots:
                self._slots.remove(slot_id)

    async def ping(self) -> float:
        if not await self.active():
            raise CommunicationError("LTX is not running")

        req = Requests.ping()
        start_t = time.monotonic()
        replies = await self._send_requests([req])

        return (replies[req][0] * 1e-9) - start_t

    async def communicate(self, iobuffer: Optional[IOBuffer] = None) -> None:
        if await self.active():
            raise CommunicationError("LTX is already running")

        self._ltx = LTX(self._infile, self._outfile)

        try:
            await self._ltx.connect()
        except LTXError as err:
            raise CommunicationError(err) from err

        await self._send_requests([Requests.version()])

    async def run_command(
        self,
        command: str,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        iobuffer: Optional[IOBuffer] = None,
    ) -> Optional[Dict[str, Any]]:
        if not command:
            raise ValueError("command is empty")

        if not await self.active():
            raise CommunicationError("LTX is not running")

        self._logger.info("Running command: %s", repr(command))

        slot_id = await self._reserve_slot()
        ret = None

        try:
            start_t = time.monotonic()

            requests: List[Request] = []
            if cwd:
                requests.append(Requests.cwd(slot_id, cwd))

            if env:
                for key, value in env.items():
                    requests.append(Requests.env(slot_id, key, value))

            async def _stdout_coro(data: str) -> None:
                if iobuffer:
                    await iobuffer.write(data)

            exec_req = Requests.execute(slot_id, command, stdout_coro=_stdout_coro)

            requests.append(exec_req)
            replies = await self._send_requests(requests)
            reply = replies[exec_req]

            ret = {
                "command": command,
                "stdout": reply[3],
                "exec_time": (reply[0] * 1e-9) - start_t,
                "returncode": reply[2],
            }

            self._logger.debug(ret)
        finally:
            await self._release_slot(slot_id)

        self._logger.info("Command executed")

        return ret

    async def fetch_file(self, target_path: str) -> bytes:
        if not target_path:
            raise ValueError("target path is empty")

        if not await self.active():
            raise CommunicationError("LTX connection is not present")

        async with self._fetch_lock:
            req = Requests.get_file(target_path)
            replies = await self._send_requests([req])
            reply = replies[req]

            return reply[1]



"""
.. module:: plugin
    :platform: Linux
    :synopsis: generic plugin handling

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import importlib
import importlib.util
import inspect
import os
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    TypeVar,
)

_Self = TypeVar("_Self", bound="Plugin")


class Plugin:
    """
    Generic plugin definition.
    """

    _name = ""

    def setup(self, **kwargs: Dict[str, Any]) -> None:
        """
        Initialize plugin using configuration dictionary.

        :param kwargs: SUT configuration.
        :type kwargs: dict
        """
        raise NotImplementedError()

    @property
    def config_help(self) -> Dict[str, str]:
        """
        Associate each configuration option with a help message.
        This is used by the main menu application to generate --help message.

        :returns: Dictionary that associates options to help messages.
        :rtype: dict
        """
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """
        :return: Unique name identifier of the plugin.
        :rtype: str
        """
        return self._name

    def clone(self, name: str) -> _Self:
        """
        Copy plugin and return a new instance with a given name.
        Make sure that name is unique, so external modules can
        recognize it.

        :param name: Unique identifier string given to the plugin.
        :type name: str
        :return: Cloned plugin.
        :rtype: Plugin
        """
        obj = self.__class__()
        obj._name = name

        # pyrefly:ignore[bad-return]
        return obj


def discover(mytype: type, folder: str) -> List[Plugin]:
    """
    Discover mytype implementations inside a specific folder.

    :param mtype: Type of the object we are searching for.
    :type mtype: type
    :param folder: Folder where to search for mtype.
    :type folder: str
    :return: List of discovered plugins.
    :rtype: list(Plugin)
    """
    if not folder or not os.path.isdir(folder):
        raise ValueError("Discover folder doesn't exist")

    loaded_obj = []

    # use pathlib for more efficient iteration
    folder_path = Path(folder)

    for py_file in folder_path.glob("*.py"):
        if not py_file.is_file():
            continue

        spec = importlib.util.spec_from_file_location("obj", str(py_file))
        if not spec or not spec.loader:
            continue

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        module_name = module.__name__

        members = inspect.getmembers(module, inspect.isclass)
        for _, klass in members:
            if (
                klass.__module__ != module_name
                or klass is mytype
                or klass in loaded_obj
            ):
                continue

            if issubclass(klass, mytype):
                loaded_obj.append(klass())

    if loaded_obj:
        loaded_obj.sort(key=lambda x: x.name)

    return loaded_obj
